import logging
import sys
import contextlib
from typing import Optional, Any
import jax
import jax.numpy as jnp
import numpy
import itertools
import torch
import torch.distributed._functional_collectives
import torch.func
import torch.utils._mode_utils as mode_utils
import torch.utils._python_dispatch as torch_dispatch
import torch.utils._pytree as torch_pytree
from torchax.view import View
from torchax import config
from torchax.ops import mappings, ops_registry
from torchax import amp
from jax.experimental import mutable_array

logger = logging.getLogger(__name__)


class OperatorNotFound(Exception):
  pass


def wrap(jaxarray):
  return torch_pytree.tree_map_only(jnp.ndarray, Tensor, jaxarray)


def unwrap(torchtensors):
  return torch_pytree.tree_map_only(Tensor, lambda x: x._elem, torchtensors)


@contextlib.contextmanager
def log_nested(env, message):
  if env.config.debug_print_each_op:
    print((" " * log_nested.level) + message, file=sys.stderr)
  log_nested.level += 1
  yield
  log_nested.level -= 1


log_nested.level = 0


class Tensor(torch.Tensor):

  @staticmethod
  def __new__(cls, elem, env):
    dtype = mappings.j2t_dtype(elem.dtype)
    shape = list(elem.shape)
    for i, s in enumerate(shape):
      if not isinstance(s, int):
        shape[i] = 1
    if dtype is None:
      dtype = torch.float32
    return torch.Tensor._make_wrapper_subclass(
        cls,
        shape,
        dtype=dtype,
        device="meta",
        requires_grad=False,
    )

  def __init__(self, elem: jax.Array, env: "Environment"):
    super().__init__()
    self._elem = elem
    self._env = env

  def __str__(self):
    return "Tensor({} {})".format(str(type(self._elem)), str(self._elem))

  __repr__ = __str__

  def __jax_array__(self):
    return self._elem

  @property
  def shape(self):
    return torch.Size(self._elem.shape)

  @property
  def ndim(self):
    return len(self._elem.shape)

  def flatten(self, start_dim=0, end_dim=-1):
    if end_dim == -1:
      end_dim = self.ndim
    new_shape = (
        self._elem.shape[:start_dim] + (-1,) + self._elem.shape[end_dim + 1:])
    new_elem = jnp.reshape(self._elem, new_shape)
    return Tensor(new_elem, self._env)
    # return torch.reshape(self, new_shape)

  def __setitem__(self, key, val):
    key, val = self._env.t2j_iso((key, val))
    self._elem = self._elem.at[key].set(val)

  def type_as(self, other):
    self._elem = self._elem.astype(other._elem.dtype)
    return self

  __torch_function__ = torch._C._disabled_torch_function_impl

  @classmethod
  def __torch_dispatch__(cls, func, types, args=(), kwargs=None):
    # TODO(hanq): figure out why is dispatch mode not sufficient
    if func == torch.ops._c10d_functional.wait_tensor.default:
      return args[0]._env.dispatch(func, types, args, kwargs)
    raise AssertionError(
        'torchax Tensors can only do math within the torchax environment.'
        'Please wrap your code with `with torchax.default_env()` or '
        'call torchax.enable_globally() before.')

  def detach(self):
    return Tensor(jax.lax.stop_gradient(self.jax()), self._env)

  def numpy(self) -> numpy.ndarray:
    import numpy as np

    return np.array(self._elem)

  def jax(self) -> jax.Array:
    return self._elem

  def torch(self) -> torch.Tensor:
    return self._env.j2t_copy(self.jax())

  @property
  def dtype(self):
    return mappings.j2t_dtype(self._elem.dtype)

  def dim(self):
    return self.ndim

  @property
  def device(self):
    return torch.device("jax:0")

  @property
  def jax_device(self):
    return self._elem.device

  @property
  def data(self):
    logger.warning(
        "In-place to .data modifications still results a copy on TPU")
    return self

  @data.setter
  def data(self, other):
    if isinstance(other, Tensor):
      self._elem = other._elem

  def apply_jax(self, jax_function, *args, **kwargs):
    # Call a jax function on _elem
    res = jax_function(self._elem, *args, **kwargs)
    return self._env.j2t_iso(res)

  def apply_jax_(self, jax_function, *args, **kwargs):
    self._elem = jax_function(self._elem, *args, **kwargs)
    return self

  def tolist(self):
    return self._elem.tolist()

  def shard_(self, sharding):
    self.apply_jax_(jax.lax.with_sharding_constraint, sharding)


def debug_accuracy(func, args, kwargs, current_output):
  args_torch, kwargs_torch, out_torch = torch_pytree.tree_map_only(
      torch.Tensor, lambda x: x.torch(), (args, kwargs, current_output))

  with mode_utils.no_dispatch(), torch._C.DisableTorchFunction():
    if "device" in kwargs_torch:
      kwargs_torch["device"] = "cpu"  # do the torch native for comparison
    expected_out = func(*args_torch, **kwargs_torch)

  flattened_current_out, _ = torch_pytree.tree_flatten(out_torch)
  flattened_expected_out, _ = torch_pytree.tree_flatten(expected_out)

  for ex, real in zip(flattened_expected_out, flattened_current_out):
    if isinstance(ex, torch.Tensor) and ex.dtype != real.dtype:
      ex = ex.to(real.dtype)
    try:
      if isinstance(ex, torch.Tensor) and not torch.allclose(
          ex, real, atol=1e-3, equal_nan=True):
        import pdb

        pdb.set_trace()
    except:
      import pdb

      pdb.set_trace()

  return True


def _make_debug_msg(is_dispatch, log_args, func, args, kwargs):

  def _display(a):
    if isinstance(a, torch.Tensor):
      return f"Tensor of {type(a)}: {a.dtype}{a.shape}"
    elif isinstance(a, jax.Array):
      return f"Jax Array of {type(a)}: {a.dtype}{a.shape}"
    else:
      return str(a)

  kwargs = kwargs or {}
  title = "DISPATCH" if is_dispatch else "FUNCTION"
  args_msg = "args: " + ",".join(_display(a) for a in args) if log_args else ""
  kwargs_msg = ("kwargs: " +
                ",".join(f"{key}: {_display(a)}" for key, a in kwargs.items())
                if log_args else "")
  return f"{title}: {_name_of_func(func)} {args_msg} ~ {kwargs_msg}"


class XLAFunctionMode(torch.overrides.TorchFunctionMode):
  """Context manager that dispatches torch function calls to JAX."""

  def __init__(self, env):
    self.env = env

  def __torch_function__(self,
                         func,
                         types,
                         args=(),
                         kwargs=None) -> torch.Tensor:
    message = f"FUNCTION: {_name_of_func(func)}"
    if self.env.config.debug_print_each_op_operands:
      message = message + "f"
    message = _make_debug_msg(False,
                              self.env.config.debug_print_each_op_operands,
                              func, args, kwargs)
    with log_nested(self.env, message):
      try:
        return self.env.dispatch(func, types, args, kwargs)
      except OperatorNotFound:
        pass
      if _name_of_func(func) in (
          "rot90"):  # skip rot90 with k%4==0 due to no change
        if len(args) >= 2 and type(args[1]) == int:
          if (args[1]) % 4 == 0:
            return args[0]
      return func(*args, **(kwargs or {}))


class XLADispatchMode(torch_dispatch.TorchDispatchMode):

  def __init__(self, env):
    self.env = env

  def __torch_dispatch__(self, func, types, args=(), kwargs=None):
    message = _make_debug_msg(True,
                              self.env.config.debug_print_each_op_operands,
                              func, args, kwargs)
    with log_nested(self.env, message):
      if isinstance(func, torch._ops.OpOverloadPacket):
        with self:
          return func(*args, **kwargs)
      # Only functions under these namespaces will be intercepted
      if func.namespace not in (
          "aten",
          "_c10d_functional",
          "torchvision",
          "xla",
      ):
        return func(*args, **kwargs)
      return self.env.dispatch(func, types, args, kwargs)


def _name_of_func(func):
  if hasattr(func, "name"):
    return func.name()
  return func.__name__


# Constructors that don't take other tensor as input
TENSOR_CONSTRUCTORS = {
    torch.ones,
    torch.zeros,
    torch.empty,
    torch.empty_strided,
    torch.tensor,
    torch.arange,
    torch.eye,
    torch.randn,
    torch.rand,
    torch.randint,
    torch.full,
    torch.as_tensor,
}

# TODO(wen): use existing types, either from torch or jax
SUPPORTED_JAX_PLATFROM = ["cpu", "tpu"]


class Environment(contextlib.ContextDecorator):
  """This class holds a set of configurations and "globals" needed

  for executing torch program using jax.
  Things included so far:

  op registry
  PRNGKey
  Configs

  Also helper functions to manipulate those.
  """

  def __init__(self, configuration=None):
    self._function_mode = XLAFunctionMode(self)
    self._dispatch_mode = XLADispatchMode(self)

    # name is torch callable
    self._ops = {}
    self._decomps = {}

    self.load_ops()

    self._mesh = None
    self.config = configuration or config.Configuration()

    self._manually_entered = False
    self.enabled = False

    self._prng_key = mutable_array(
        jax.random.key(torch.initial_seed() % (1 << 63)))
    self.autocast_dtype = None
    self._target_device = jax.local_devices()[0].platform

  @property
  def target_device(self):
    return self._target_device

  @target_device.setter
  def target_device(self, device: str):
    self._target_device = device.lower()

  def manual_seed(self, key):
    self._prng_key = mutable_array(jax.random.key(key))

  @property
  def prng_key(self):
    return self._prng_key[...]

  def get_as_jax_device(self, device: Any):
    if device is None:
      device = torch.get_default_device()

    if isinstance(device, torch.device):
      device = str(device)

    if not self.config.use_torch_native_for_cpu_tensor and device.startswith(
        "cpu"):
      return jax.devices("cpu")[0]

    if self.config.treat_cuda_as_jax_device and device.startswith("cuda"):
      return jax.local_devices()[0]

    if device.startswith("xla"):
      return jax.local_devices()[0]

    # TODO (wen): jax is NOT a device type,
    # once we can register more than one backend, revisit
    if device.startswith("jax"):
      match self.target_device:
        case "cpu":
          return jax.devices("cpu")[0]
        case "tpu":
          return jax.devices("tpu")[0]
        case _:
          raise AttributeError(
              f"Cannot handle env.target_device {self.target_device}")

    return None  # fallback to torch

  def load_ops(self):
    from torchax.ops import jaten, jtorch, jc10d, jtorchvision_nms

    for k, v in itertools.chain(ops_registry.all_aten_ops.items(),
                                ops_registry.all_torch_functions.items()):
      if v.is_jax_function:
        self._ops[k] = v
      else:
        self._decomps[k] = v

    from torchax.decompositions import DECOMPOSITIONS, MUTABLE_DECOMPOSITION

    for k, v in DECOMPOSITIONS.items():
      if k not in self._decomps:
        self._decomps[k] = ops_registry.Operator(
            k,
            v,
            is_jax_function=False,
            is_user_defined=False,
            needs_env=False,
            is_view_op=k in MUTABLE_DECOMPOSITION,
        )

  def _get_op_or_decomp(self, func):

    def _get_from_dict(op_dict, op):
      op = op_dict.get(func)
      if op is None and isinstance(func, torch._ops.OpOverloadPacket):
        op = op_dict.get(func.default)
      if op is None and isinstance(func, torch._ops.OpOverload):
        op = op_dict.get(func.overloadpacket)
      return op

    op = _get_from_dict(self._ops, func)

    if op is None:
      # fallback to decompose
      op = _get_from_dict(self._decomps, func)

    if op is None:
      raise OperatorNotFound(
          f"Operator with name {_name_of_func(func)} has no lowering")

    return op

  def _to_copy(self, the_tensor, new_dtype, new_device):
    if isinstance(the_tensor, View):
      the_tensor = the_tensor.torch()

    if isinstance(the_tensor, Tensor):

      arr = the_tensor.jax()

      if new_dtype is not None and new_dtype != arr.dtype:
        arr = arr.astype(mappings.t2j_dtype(new_dtype))

      if new_device is not None:
        match str(new_device).lower():
          case "cpu":
            # converting to a non-jax device: let torch native handle it
            torch_tensor = self.j2t_copy(arr) if isinstance(the_tensor,
                                                            Tensor) else arr
            with mode_utils.no_dispatch(), torch._C.DisableTorchFunction():
              return torch_tensor.to(new_device)
          case "jax":
            # move torchax.tensor / jax tensor between devices
            # I don't know ifgit  this will work after the model is jitted
            if self.target_device != the_tensor.jax_device.platform:
              arr = jax.device_put(the_tensor.jax(),
                                   jax.devices(self.target_device)[0])
              return Tensor(arr, self)
          case _:
            logging.error(f"torchax.Tenosr cannot handle device {new_device}")

    else:
      if new_dtype is not None and new_dtype != the_tensor.dtype:
        with mode_utils.no_dispatch(), torch._C.DisableTorchFunction():
          the_tensor = the_tensor.to(new_dtype)

      if new_device is None:  ## device is None means don't change device
        return the_tensor

      jax_device = self.get_as_jax_device(new_device)
      if jax_device:
        arr = self.t2j_copy(the_tensor)
        arr = jax.device_put(arr, jax_device)
      else:
        with mode_utils.no_dispatch(), torch._C.DisableTorchFunction():
          return the_tensor.to(new_device)

    return Tensor(arr, self)

  def get_and_rotate_prng_key(self,
                              generator: Optional[torch.Generator] = None):
    if generator is not None:
      with mode_utils.no_dispatch(), torch._C.DisableTorchFunction():
        self._prng_key[...] = jax.random.key(generator.initial_seed() % (2**63))
    old_key = self._prng_key[...]
    new_prng_key, next_key = jax.random.split(old_key)
    self._prng_key[...] = new_prng_key
    return next_key

  def _handle_tensor_constructor(self, func, args, kwargs):
    device = kwargs.get("device")
    jax_device = self.get_as_jax_device(device)
    # TODO(qihqi) figure out better ways for device propagation
    if not self._manually_entered and jax_device is None:
      # let torch handle it
      with mode_utils.no_dispatch(), torch._C.DisableTorchFunction():
        return func(*args, **kwargs)
    with jax.default_device(jax_device):
      requires_grad = kwargs.get("requires_grad", False)
      op = self._get_op_or_decomp(func)
      res = op.func(*args, **kwargs)
      if isinstance(res, jax.Array):
        res = Tensor(res, self)
      if requires_grad:
        res.requires_grad = True
      return res

  def _torch_Tensor_to(self, args, kwargs):
    the_tensor = args[0]
    args = args[1:]
    if len(args) >= 1 and isinstance(args[0], torch.Tensor):
      dtype = args[0].dtype
      device = args[0].device
      return self._to_copy(the_tensor, dtype, device)
    device = kwargs.get("device")
    dtype = kwargs.get("dtype")
    # args like pin_memory etc that we will ignore
    args = list(filter(lambda x: not isinstance(x, bool), args))
    if len(args) >= 2:
      device, dtype, *_ = args
    elif len(args) == 1 and isinstance(args[0], torch.dtype):
      dtype = args[0]
    elif len(args) == 1:
      device = args[0]
    return self._to_copy(the_tensor, dtype, device)

  def dispatch(self, func, types, args, kwargs):
    kwargs = kwargs or {}
    if func in TENSOR_CONSTRUCTORS:
      return self._handle_tensor_constructor(func, args, kwargs)
    if func in (
        torch.Tensor.to,
        torch.ops.aten.lift_fresh.default,
        torch.ops.aten._to_copy,
        torch.ops.aten._to_copy.default,
    ):
      return self._torch_Tensor_to(args, kwargs)

    # If the func doesn't act on Tensor, and is not a tensor constructor,
    # We should skip and let torch handle it.

    tensor_args = [
        t for t in torch_pytree.tree_flatten(args)[0]
        if isinstance(t, torch.Tensor)
    ]

    def is_not_torchax_tensor(x):
      return not isinstance(x, Tensor) and not isinstance(x, View)

    if tensor_args and all(is_not_torchax_tensor(t) for t in tensor_args):
      res = func(*args, **kwargs)
      return res

    with jax.named_scope(_name_of_func(func)):
      op = self._get_op_or_decomp(func)

      old_args, old_kwargs = args, kwargs
      with self._dispatch_mode:
        args, kwargs = torch_pytree.tree_map_only(
            torch.distributed._functional_collectives.AsyncCollectiveTensor,
            torch.distributed._functional_collectives.wait_tensor,
            (args, kwargs),
        )

      try:
        if not op.is_view_op:
          args, kwargs = self.v2t_iso((args, kwargs))

        with self:
          if self.autocast_dtype is not None:
            autocast_policy = amp.autocast_policy.get(func)
            if autocast_policy is not None:
              args, kwargs = amp.execute_policy(autocast_policy, args, kwargs,
                                                self.autocast_dtype)

        if op.is_jax_function:
          args, kwargs = self.t2j_iso((args, kwargs))
      except AssertionError:
        if self.config.debug_mixed_tensor:
          breakpoint()
        else:
          raise

      if op.needs_env:
        kwargs["env"] = self

      if op.is_jax_function:
        res = op.func(*args, **kwargs)
      else:
        # enable dispatch mode because this op could be a composite autograd op
        # meaning, it will decompose in C++
        with self._dispatch_mode:
          res = op.func(*args, **kwargs)

      if op.is_jax_function:
        res = self.j2t_iso(res)

      if self.config.force_materialize_views and isinstance(res, View):
        res = res.torch()

      if self.config.debug_accuracy_for_each_op:
        debug_accuracy(func, old_args, old_kwargs, res)
      return res

  def enable_torch_modes(self):
    self._dispatch_mode.__enter__()
    self._function_mode.__enter__()
    self.enabled = True

  def disable_torch_modes(self, *exc):
    if not exc:
      exc = (None, None, None)
    self._function_mode.__exit__(*exc)
    self._dispatch_mode.__exit__(*exc)
    self.enabled = False

  def __enter__(self):
    self.enable_torch_modes()
    self._manually_entered = True
    return self

  def __exit__(self, *exc):
    self._manually_entered = False
    self.disable_torch_modes(*exc)

  def _move_one_value(self, val):
    if isinstance(val, torch.nn.Module):
      with self:
        return val.to("jax")
    if isinstance(val, Tensor):
      return val
    if isinstance(val, torch.Tensor):
      return Tensor(self.t2j_copy(val), self)
    return val

  def to_xla(self, torchvalues):
    # tensors are torch.Tensors (not XLATensor)
    res = torch_pytree.tree_map(self._move_one_value, torchvalues)
    return res

  def t2j_iso(self, torchtensors):
    """Convert torchax Tensor to jax array.
    
    This function will not copy, will just unwrap the inner jax array out.
    Note: iso is short for "isomorphic"
    """

    def to_jax(x):
      if isinstance(
          x, torch.distributed._functional_collectives.AsyncCollectiveTensor):
        x = x.wait()
      assert isinstance(x, Tensor) or isinstance(x, View), (
          f"Expect a Tensor or a View but got {type(x)}; usually this means there is a mixed math between XLATensor and torch.Tensor"
      )
      return x.jax()

    res = torch_pytree.tree_map_only(torch.Tensor, to_jax, torchtensors)
    return res

  def v2t_iso(self, views):

    def to_tensor(x):
      if isinstance(x, View):
        return x.torch()
      return x

    res = torch_pytree.tree_map_only(View, to_tensor, views)
    return res

  def j2t_iso(self, jaxarray):
    """Convert jax array to torchax Tensor.
    
    This function will not copy, will just wrap the jax array with a torchax Tensor
    Note: iso is short for "isomorphic"
    """
    return torch_pytree.tree_map_only(jax.Array, lambda x: Tensor(x, self),
                                      jaxarray)

  def j2t_copy(self, args):
    """Convert torch.Tensor in cpu to a jax array
    
    This might involves copying the data (depending if dlpack is enabled)
    """
    return torch_pytree.tree_map_only(
        jax.Array,
        lambda x: mappings.j2t(x, self.config.use_dlpack_for_data_conversion),
        args)

  def t2j_copy(self, args):
    """Convert jax array to torch.Tensor in cpu.
    
    This might involves copying the data (depending if dlpack is enabled)
    """
    return torch_pytree.tree_map_only(
        torch.Tensor,
        lambda x: mappings.t2j(x, self.config.use_dlpack_for_data_conversion),
        args)

  def override_op_definition(self, op_to_override, op_impl):
    self._ops[op_to_override] = ops_registry.Operator(
        op_to_override,
        op_impl,
        is_jax_function=False,
        is_user_defined=True,
        needs_env=False,
    )



import unittest
from torchax.util import partition, merge


# Helper predicate functions for testing partition
def is_even(n):
  return isinstance(n, int) and n % 2 == 0


def is_positive(n):
  return isinstance(n, (int, float)) and n > 0


def is_string(s):
  return isinstance(s, str)


class TestListUtils(unittest.TestCase):

  # --- Tests for partition ---

  def test_partition_empty_list(self):
    """Test partition with an empty list."""
    self.assertEqual(partition([], is_even), ([], []))

  def test_partition_even_odd(self):
    """Test partitioning numbers into even and odd."""
    nums = [1, 2, 3, 4, 5, 6]
    expected_truthy = [None, 2, None, 4, None, 6]
    expected_falsy = [1, None, 3, None, 5, None]
    self.assertEqual(
        partition(nums, is_even), (expected_truthy, expected_falsy))

  def test_partition_all_true(self):
    """Test partition when the predicate is always true."""
    evens = [2, 4, 6, 8]
    expected_truthy = [2, 4, 6, 8]
    expected_falsy = [None, None, None, None]
    self.assertEqual(
        partition(evens, is_even), (expected_truthy, expected_falsy))

  def test_partition_all_false(self):
    """Test partition when the predicate is always false."""
    odds = [1, 3, 5, 7]
    expected_truthy = [None, None, None, None]
    expected_falsy = [1, 3, 5, 7]
    self.assertEqual(
        partition(odds, is_even), (expected_truthy, expected_falsy))

  def test_partition_mixed_types(self):
    """Test partition with a list of mixed types."""
    mixed = [1, "hello", 2.5, "world", 3, None]
    # Using is_string as the predicate
    expected_truthy = [None, "hello", None, "world", None, None]
    expected_falsy = [1, None, 2.5, None, 3,
                      None]  # Note: None itself is not a string
    self.assertEqual(
        partition(mixed, is_string), (expected_truthy, expected_falsy))

  def test_partition_with_lambda(self):
    """Test partition using a lambda function as the predicate."""
    nums = [-2, -1, 0, 1, 2]
    expected_truthy = [None, None, None, 1, 2]
    expected_falsy = [-2, -1, 0, None, None]
    self.assertEqual(
        partition(nums, lambda x: isinstance(x, int) and x > 0),
        (expected_truthy, expected_falsy))

  # --- Tests for merge ---

  def test_merge_empty_lists(self):
    """Test merge with empty lists."""
    self.assertEqual(merge([], []), [])

  def test_merge_basic(self):
    """Test basic merging with None values in the first list."""
    list1 = [1, None, 3, None, 5]
    list2 = [None, 2, None, 4, None]
    expected = [1, 2, 3, 4, 5]
    self.assertEqual(merge(list1, list2), expected)

  def test_merge_no_none_in_list1(self):
    """Test merge when list1 has no None values."""
    list1 = ['a', 'b', 'c']
    list2 = [1, 2, 3]
    expected = ['a', 'b', 'c']  # Should be identical to list1
    self.assertEqual(merge(list1, list2), expected)

  def test_merge_all_none_in_list1(self):
    """Test merge when list1 contains only None."""
    list1 = [None, None, None]
    list2 = ['x', 'y', 'z']
    expected = ['x', 'y', 'z']  # Should be identical to list2
    self.assertEqual(merge(list1, list2), expected)

  def test_merge_mixed_types(self):
    """Test merge with mixed data types."""
    list1 = [1, None, "hello", None]
    list2 = [None, 2.5, None, True]
    expected = [1, 2.5, "hello", True]
    self.assertEqual(merge(list1, list2), expected)

  def test_merge_unequal_lengths(self):
    """Test that merge raises AssertionError for lists of unequal length."""
    list1 = [1, 2, 3]
    list2 = [4, 5]
    # Use assertRaises as a context manager
    with self.assertRaises(AssertionError) as cm:
      merge(list1, list2)

    list3 = [6, 7]
    list4 = [8, 9, 10]
    with self.assertRaises(AssertionError):
      merge(list3, list4)  # No need to check message again if already checked


if __name__ == '__main__':
  unittest.main()  # For running from command line



from typing import Tuple
import time
import torchax
from torchax.interop import jax_view, torch_view, JittableModule
import jax
import optax
import jax
import jax.numpy as jnp
from jax.experimental import mesh_utils
import functools
from torch.utils import _pytree as pytree

Mesh = jax.sharding.Mesh
P = jax.sharding.PartitionSpec

SEQLEN = 8192
BATCH = 8
global_axis: Tuple[str, str] = ('fsdp',)
num_global_devices = jax.device_count()
num_local_devices = jax.local_device_count()
num_partitions = (num_global_devices,)
#SEQLEN = 512

import torch


def group_data(dataloader, block_size):
  """yields tuple of inputs, label with seqlen == block_size"""

  tally = 0
  inputs = []
  labels = []

  for line in dataloader:
    x, y = line['input_ids'], line['labels']
    inputs.append(x)
    labels.append(y)
    batch, seqlen = x.shape
    tally += seqlen
    if tally > block_size:
      inputs_stacked = torch.concat(inputs, dim=-1)
      inputs_stacked.resize_((batch, block_size))
      labels_stacked = torch.concat(labels, dim=-1)
      labels_stacked.resize_((batch, block_size))
      yield inputs_stacked, labels_stacked
      tally = 0
      inputs = []
      labels = []


def sharded_device_put(tensor, sharding):
  if isinstance(tensor, tuple):
    return tuple(sharded_device_put(t, sharding) for t in tensor)

  if num_global_devices == num_local_devices:
    return jax.device_put(tensor, sharding)

  shape = tensor.shape
  x_split = [
      jax.device_put(tensor[i], device)
      for device, i in sharding.addressable_devices_indices_map(shape).items()
  ]
  return jax.make_array_from_single_device_arrays(shape, sharding, x_split)


class FSDPv2(torch.nn.Module):

  def __init__(self, mod):
    super().__init__()
    self.mod = mod
    self.mesh = jax.sharding.Mesh(
        mesh_utils.create_device_mesh(num_partitions),
        axis_names=global_axis,
    )
    self.sharding = jax.sharding.NamedSharding(self.mesh, P(*global_axis))

  def forward(self, *args):
    args = list(args)
    args[0] = self.shard(args[0])
    res = self.mod(*args)
    return self.shard(res)

  def shard(self, x):
    return torchax.interop.call_jax(
        jax.lax.with_sharding_constraint,
        x,
        self.sharding,
    )


def print_shapes(pyt):
  for p in pytree.tree_flatten(pyt)[0]:
    if hasattr(p, 'shape'):
      print(p.shape, p.dtype)


class JaxTrainer:

  def __init__(self, use_fori):
    self.use_fori = use_fori
    self.mesh = jax.sharding.Mesh(
        mesh_utils.create_device_mesh(num_partitions),
        axis_names=global_axis,
    )
    self.x_sharding = jax.sharding.NamedSharding(self.mesh, P(global_axis))
    self.y_sharding = jax.sharding.NamedSharding(self.mesh, P(*global_axis))
    self.replicated = jax.sharding.NamedSharding(self.mesh, P())

  def torch_opt_to_jax_opt(self, torch_opt):
    # TODO: Can convert optimizer instead of using a jax one
    return optax.adamw(0.01)

  def fit_model_fori(self, gpt_mod, data_loader):
    xla_env = torchax.default_env()
    jax.config.update('jax_enable_x64', False)
    xla_env._mesh = self.mesh
    xla_env.use_flash_attention = True

    weights = gpt_mod.weights

    jax_params = {}
    for k, v in weights.items():
      sharding = self.y_sharding if k == 'block' else self.x_sharding
      print(k, sharding)
      jax_params[k] = self._shard_fsdp_style(v, sharding)

    print('ALL weights ===')
    for x in jax.tree_util.tree_flatten(jax_params)[0]:
      print(x.shape, x.sharding)
    print(' ===')

    @jax.checkpoint
    def loss(jax_params, data):
      data = jax.lax.with_sharding_constraint(data, self.x_sharding)  # fsdpv2
      x, y = data
      res = torchax.interop.call_torch(gpt_mod.forward_with_weights, jax_params,
                                       x)
      res = jax.lax.with_sharding_constraint(res, self.x_sharding)
      return jnp.mean(
          optax.losses.softmax_cross_entropy_with_integer_labels(res, y))

    grad_fn = jax.value_and_grad(loss)
    jax_optimizer = optax.adamw(0.01)
    opt_state = jax_optimizer.init(jax_params)

    @functools.partial(jax.jit, donate_argnums=(0, 1))
    def step(jax_weights, opt_state, data):
      with jax.named_scope('compute_gradient'):
        loss, gradient = grad_fn(jax_weights, data)
      with jax.named_scope("optimizer_updates"):
        updates, opt_state = jax_optimizer.update(gradient, opt_state,
                                                  jax_weights)
        jax_weights = optax.apply_updates(jax_weights, updates)
      return loss, jax_weights, opt_state

    print('Start compiling')
    start = time.perf_counter()
    lowered = step.lower(
        jax_params,
        opt_state,
        (jax.ShapeDtypeStruct(
            (BATCH, SEQLEN), jnp.dtype('int32'), sharding=self.x_sharding),
         jax.ShapeDtypeStruct(
             (BATCH, SEQLEN), jnp.dtype('int32'), sharding=self.x_sharding)),
    )
    # print(lowered.as_text())
    print('program size:', len(lowered.as_text()) / 1e6, 'm chars')
    step_compiled = lowered.compile()
    end = time.perf_counter()
    print('End compiling', end - start)
    compile_time = end - start

    for co in step_compiled.cost_analysis():
      print('flops counter:', co['flops'])

    s = time.perf_counter()
    jax.profiler.start_trace('/tmp/tensorboard')
    print('start training')
    min_loop_time = 10000
    for i, item in enumerate(group_data(data_loader, SEQLEN)):
      inputs, labels = sharded_device_put(
          jax_view(xla_env.to_xla(item)), self.x_sharding)
      print('INPUT shape', inputs.shape)

      step_start = time.perf_counter()
      loss, jax_params, opt_state = step_compiled(jax_params, opt_state,
                                                  (inputs, labels))
      jax.block_until_ready((loss, jax_params))
      step_end = time.perf_counter()
      print(i, 'loss', loss, 'step latency: ', step_end - step_start)
      min_loop_time = min(min_loop_time, step_end - step_start)
      print('======')
      if i >= 3:
        break
    jax.profiler.stop_trace()
    return min_loop_time, compile_time

  def _shard_fsdp_style(self, state_dict, sharding=None):
    if sharding is None:
      sharding = self.x_sharding

    def move_one_tensor(x):
      env = torchax.default_env()
      jval = env.t2j_copy(x)
      return sharded_device_put(jval, sharding)

    if isinstance(state_dict, torch.Tensor):
      return move_one_tensor(state_dict)
    res = {}
    for k, v in sorted(state_dict.items()):
      res[k] = move_one_tensor(v)
    return res

  def fit(self, lightning_mod, data_loader):

    xla_env = torchax.default_env()
    jax.config.update('jax_enable_x64', False)
    xla_env._mesh = self.mesh
    xla_env.use_flash_attention = True

    jittable_mod = JittableModule(lightning_mod)
    jax_params = self._shard_fsdp_style(jittable_mod.params)
    jax_buffers = self._shard_fsdp_style(jittable_mod.buffers)

    @jax.checkpoint
    def lightning_mod_loss(weights: jax.Array, buffers: jax.Array,
                           data: jax.Array, batch_id):
      """returns loss"""
      with jax.named_scope("Computing_loss"):
        weights, buffers, data = torch_view((weights, buffers, data))
        # NOTE: these is needed because the original model
        # did not register those as persistent buffer
        with xla_env:
          loss = jittable_mod.functional_call('training_step', weights, buffers,
                                              data, batch_id)
        return jax_view(loss)

    jax_optimizer = self.torch_opt_to_jax_opt(
        lightning_mod.configure_optimizers())

    opt_state = jax_optimizer.init(jax_params)
    grad_fn = jax.value_and_grad(lightning_mod_loss)

    opt_state_sharding = jax.tree_util.tree_map(lambda p: p.sharding, opt_state)

    print('Begining training')

    # NOTE: explicitly set sharding so the sharding of opt_state wont change
    # if it changes, it would trigger recompile
    @functools.partial(
        jax.jit,
        donate_argnums=(0, 2),
        #in_shardings=(self.x_sharding, self.x_sharding, opt_state_sharding, self.x_sharding, self.replicated),
        #out_shardings=(self.replicated, self.x_sharding, opt_state_sharding),
    )
    def step(jax_weights, jax_buffers, optimizer_state, xla_data, bid):
      print('Tracing inside of step')
      with jax.named_scope("Computing_loss_and_grad"):
        loss, grads = grad_fn(jax_weights, jax_buffers, xla_data, bid)
      with jax.named_scope("optimizer_updates"):
        updates, opt_state = jax_optimizer.update(grads, optimizer_state,
                                                  jax_weights)
        jax_weights = optax.apply_updates(jax_weights, updates)
      return loss, jax_weights, opt_state

    total_param_size = 0
    for k, v in jax_params.items():
      total_param_size += v.size

    print('Total number of params: ', total_param_size)
    # print(jax.jit(jax.grad(lightning_mod_loss)).lower(
    #     jax_params, jax_buffers,
    #     (jax.ShapeDtypeStruct((8, SEQLEN), jnp.dtype('int32')),
    #      jax.ShapeDtypeStruct((8, SEQLEN), jnp.dtype('int32'))),
    #     0
    # ).as_text())

    print('Start compiling')
    start = time.perf_counter()
    lowered = step.lower(
        jax_params, jax_buffers, opt_state,
        (jax.ShapeDtypeStruct(
            (8, SEQLEN), jnp.dtype('int32'), sharding=self.x_sharding),
         jax.ShapeDtypeStruct(
             (8, SEQLEN), jnp.dtype('int32'), sharding=self.x_sharding)), 0)
    # print(lowered.as_text())
    print('program size:', len(lowered.as_text()) / 1e6, 'm chars')
    step_compiled = lowered.compile()
    end = time.perf_counter()
    compile_time = end - start
    print('End compiling', compile_time)

    for co in step_compiled.cost_analysis():
      print('flops counter:', co['flops'])

    s = time.perf_counter()
    jax.profiler.start_trace('/tmp/tensorboard')
    print('start training')
    min_loop_time = 10000
    for i, item in enumerate(group_data(data_loader, SEQLEN)):
      inputs, labels = sharded_device_put(
          jax_view(xla_env.to_xla(item)), self.x_sharding)
      print('INPUT shape', inputs.shape)

      step_start = time.perf_counter()
      loss, jax_params, opt_state = step_compiled(jax_params, jax_buffers,
                                                  opt_state, (inputs, labels),
                                                  0)
      jax.block_until_ready((loss, jax_params))
      step_end = time.perf_counter()
      print(i, 'loss', loss, 'step latency: ', step_end - step_start)
      loop_time = step_end - step_start
      min_loop_time = min(min_loop_time, loop_time)
      print('======')
      if i >= 2:
        break
    jax.profiler.stop_trace()
    return min_loop_time, compile_time



import jax

# jax.grad will take f, return another callable g; such that g computes
# gradient of f. The return value of g, will have the same pytree-shape
# as the first input of f; and by default that it the arg wrt which the gradient is
# taken

# Example


def f(a_dict, b):
  return a_dict['weight'] * b + a_dict['bias']


# a_dict is a dict with 2 keys ['weights', 'bias']
# So

# jax.grad(f)(a_dict, b) will also have 2 keys, 'weights' and 'bias'

# Now, say I only want to compute the gradient of 'weight' and want to skip
# gradient computation of 'bias'. I can accomplish this by making another function
# that has the things I want to diff, and in that format, as first input:


def f2(dict_with_weight, dict_with_bias, b):
  a_dict = copy.copy(dict_with_weight)
  a_dict.update(dict_with_weight)
  res = f(a_dict, b)
  return res


# jax.grad(f2)(dict_weight_weights, dict_with_bias, b) will also have  keys, 'weights' and 'bias'

# OR, if I want to compute gradient of b as well together with gradient of a_dict:


def f3(inputs):
  a, b = inputs
  res = f(a, b)
  return res


# jax.grad(f3)((a_dict, b)) will a 2-tuple, first have 2 keys, 'weights' and 'bias', second is gradient of b

# Creating new functions can route the inputs however you'd like; then
# jax.grad will compute the gradient wrt that.abs

# Call a torch module
# Let m be a stateful module

m = Module(...)

# This is all its weights
states = m.state_dict()

# You call the module usually like this:
result = m(input1, input2)


def functional_m(weights, inputs):
  # weights is state_dict
  # inputs is a tuple of inputs, even if there is only one input to the module
  # The incantation below calls m, with weights explicitly passed in as an argument
  return torch.func.functional_call(m, weights, inputs)


# Now we can do the usual:

from torch_xla2 import interop

gradients = interop.jax_jit(functional_m)(states, (input1, input2, ...))



def all_aten_jax_ops():
  # to load the ops
  import torchax.ops.jaten  # type: ignore
  import torchax.ops.ops_registry  # type: ignore

  return {
      key: val.func
      for key, val in torchax.ops.ops_registry.all_aten_ops.items()
      if val.is_jax_function
  }



import time
import jax
from jax.tree_util import tree_map
from jax.sharding import NamedSharding
from torchax import interop

P = jax.sharding.PartitionSpec


def compile_step_func(step, weights, buffers, opt_state, args, label, mesh):
  step, weights, buffers, opt_state, args, label = interop.jax_view(
      (step, weights, buffers, opt_state, args, label))
  wshardings = tree_map(
      lambda a: a.sharding if isinstance(a, jax.Array) else None, weights)
  bshardings = tree_map(
      lambda a: a.sharding if isinstance(a, jax.Array) else None, buffers)
  oshardings = tree_map(
      lambda a: a.sharding if isinstance(a, jax.Array) else None, opt_state)
  print('Start compiling')
  start = time.perf_counter()
  lowered = jax.jit(
      step,
      donate_argnums=(0, 2),
      #in_shardings=shardings,
      out_shardings=(NamedSharding(mesh, P()), wshardings, oshardings),
  ).lower(weights, buffers, opt_state, args, label)
  #print(lowered.as_text())
  # import pdb; pdb.set_trace()
  print('program size:', len(lowered.as_text()) / 1e6, 'm chars')
  step_compiled = lowered.compile()
  end = time.perf_counter()
  print('End compiling', end - start)
  compile_time = end - start
  for co in step_compiled.cost_analysis():
    print('Flops', co['flops'])
    print('GB accessed', co['bytes accessed'] / 1e9)
  return interop.torch_view(step_compiled)



import unittest
import torch
from torch.utils import _pytree as pytree

from torchax import tensor

TestCase = unittest.TestCase
main = unittest.main


def diff_output(testcase, output1, output2, rtol, atol, equal_nan=True):
  if isinstance(output1, torch.Tensor):
    testcase.assertIsInstance(output2, torch.Tensor)
    output2_cpu = output2.detach().cpu()
    if output2_cpu.dtype != output1.dtype:
      output2_cpu = output2_cpu.to(output1.dtype)
    testcase.assertTrue(
        torch.allclose(
            output1, output2_cpu, atol=atol, rtol=rtol, equal_nan=equal_nan))
  elif isinstance(output1, (tuple, list)):
    testcase.assertIsInstance(output2, (tuple, list))
    testcase.assertEqual(len(output1), len(output2))
    for o1, o2 in zip(output1, output2):
      diff_output(testcase, o1, o2, rtol, atol)
  else:
    testcase.assertEqual(output1, output2)


def run_function_and_compare(testcase,
                             func,
                             args,
                             kwargs,
                             atol=1e-3,
                             rtol=1e-5,
                             equal_nan=True,
                             ignore_indices=False):
  with testcase.subTest("torch_eval"):
    res = func(*args, **kwargs)
    with testcase.subTest("torchax_eval"):
      args2, kwargs2 = pytree.tree_map_only(torch.Tensor, tensor.move_to_device,
                                            (args, kwargs))
      res2 = func(*args2, **kwargs2)
      res2 = pytree.tree_map_only(tensor.Tensor, lambda t: t.torch(), res2)
      with testcase.subTest("torchax_diff:" + str(atol)):
        if ignore_indices and isinstance(res, tuple) and len(res) == 2:
          diff_output(
              testcase,
              res[0],
              res2[0],
              atol=atol,
              rtol=rtol,
              equal_nan=equal_nan)
        else:
          diff_output(
              testcase, res, res2, atol=atol, rtol=rtol, equal_nan=equal_nan)


import unittest

import jax
from jax.sharding import PartitionSpec
import torch
import torchax
from torchax.mesh_util import Mesh, SingleAxisSharder


class MeshUtilTest(unittest.TestCase):

  def setUp(self):
    torchax.enable_globally()

  def test_init_module_sharded(self):

    class TestModule(torch.nn.Module):

      def __init__(self):
        super().__init__()
        self.a = torch.nn.Linear(8, 8)

    mesh = Mesh.fsdp_mesh()

    model = mesh.initialize_model_sharded(TestModule, ())
    self.assertEqual(
        len(model.a.weight.jax().addressable_shards), len(jax.devices()))
    self.assertEqual(
        len(model.a.bias.jax().addressable_shards), len(jax.devices()))

  def test_sharder_call(self):
    """Test the __call__ method produces the correct PartitionSpec."""
    sharder = SingleAxisSharder(axis_name="fsdp", axis_size=4)
    # Use a simple named tuple instead of MagicMock
    shaped_type = torch.ones((5, 8, 12))  # Middle dim divisible by 4

    spec = sharder("param_name", shaped_type)
    self.assertEqual(spec, PartitionSpec(None, "fsdp", None))

  def test_sharder_call_no_shardable(self):
    """Test __call__ when no dimension is shardable."""
    sharder = SingleAxisSharder(axis_name="fsdp", axis_size=4)
    shaped_type = torch.ones((5, 7, 11))

    with self.assertRaisesRegex(AssertionError,
                                "Unable to find a dim to shard"):
      sharder("param_name", shaped_type)


if __name__ == "__main__":
  unittest.main()



import torchax
import torchax.interop
import torch
import unittest
import jax

from test.moe import model


class TestMoe(unittest.TestCase):

  def _make_tiny_config(self):
    return model.ModelArgs(
        block_size=128,
        vocab_size=32000,
        n_layer=4,
        n_head=4,
        dim=128,
        intermediate_size=None,
        n_local_heads=-1,
        head_dim=32,
        rope_base=10000,
        norm_eps=1e-5,
        num_experts=8,
        num_activated_experts=2,
    )

  def _random_init(self, model):
    new_state_dict = {}

    for k, v in model.state_dict().items():
      new_state_dict[k] = torch.randn_like(v)

    model.load_state_dict(new_state_dict, assign=True)
    return model

  def test_moe_layer(self):
    model_args = self._make_tiny_config()

    moe_layer = model.MOEFeedForward(model_args)
    moe_layer = self._random_init(moe_layer)
    seqlen = 32
    x = torch.randn((seqlen, model_args.dim))
    res = moe_layer(x)

    env = torchax.default_env()
    model_xla = env.to_xla(moe_layer)
    x_xla = env.to_xla(x)
    with jax.default_matmul_precision('float32'):
      res_xla = model_xla(x_xla)
    res2 = res_xla.to('cpu')
    print('max diff', torch.max((res - res2).abs()))

    self.assertTrue(torch.allclose(res2, res, atol=1e-2))

    # test can jit

    def f(weights, x):
      return torch.func.functional_call(moe_layer, weights, (x,))

    fjitted = torchax.interop.jax_jit(f)
    weights_xla = env.to_xla(moe_layer.state_dict())

    print(fjitted(weights_xla, x_xla))


if __name__ == '__main__':
  unittest.main()



import contextlib
from typing import List, Dict, Any, Optional
import dataclasses
import jax
import os
import torch
from torch.utils import _pytree as pytree
from torchax import tensor
from torchax import distributed  # noqa: F401
from contextlib import contextmanager

__version__ = "0.0.5"
VERSION = __version__

__all__ = [
    'default_env',
    'extract_jax',
    'enable_globally',
]

from jax._src import xla_bridge

os.environ.setdefault('ENABLE_RUNTIME_UPTIME_TELEMETRY', '1')

# torchax:oss-begin
if getattr(jax.config, 'jax_pjrt_client_create_options', None):
  jax.config.update(
      'jax_pjrt_client_create_options',
      f'ml_framework_name:PyTorch/XLA2;ml_framework_version:{"v0.0.1"}')
# torchax:oss-end

env = None


def default_env():
  global env

  if env is None:
    env = tensor.Environment()
  return env


def extract_jax(mod: torch.nn.Module, env=None):
  """Returns a pytree of jax.ndarray and a jax callable."""
  if env is None:
    env = default_env()
  states = dict(mod.named_buffers())
  states.update(mod.named_parameters())

  states = env.t2j_copy(states)

  #@jax.jit
  def jax_func(states, inputs):
    (states, inputs) = env.j2t_iso((states, inputs))
    with env:
      res = torch.func.functional_call(mod, states, inputs, tie_weights=False)
    return env.t2j_iso(res)

  return states, jax_func


def enable_globally():
  env = default_env().enable_torch_modes()
  return env


def disable_globally():
  global env
  default_env().disable_torch_modes()


@contextlib.contextmanager
def disable_temporarily():
  prev = default_env().enabled
  if prev:
    disable_globally()
  yield ()
  if prev:
    enable_globally()


torch.utils.rename_privateuse1_backend('jax')
unsupported_dtype = [torch.quint8]
torch.utils.generate_methods_for_privateuse1_backend(
    for_tensor=True,
    for_module=True,
    for_storage=True,
    unsupported_dtype=unsupported_dtype)

import jax
import torchax.device_module

torch._register_device_module('jax', torchax.device_module)


def enable_accuracy_mode():
  jax.config.update('jax_enable_x64', True)
  jax.config.update('jax_default_matmul_precision', 'highest')
  default_env().config.internal_respect_torch_return_dtypes = True


def enable_performance_mode():
  jax.config.update('jax_enable_x64', False)
  jax.config.update('jax_default_matmul_precision', 'default')
  default_env().config.internal_respect_torch_return_dtypes = False


@dataclasses.dataclass
class CompileOptions:
  # only valid if compiling nn.Module
  methods_to_compile: List[str] = dataclasses.field(
      default_factory=lambda: ['forward'])
  jax_jit_kwargs: Dict[str, Any] = dataclasses.field(default_factory=dict)
  mode: str = 'jax'  # or dynamo or export


def compile(fn, options: Optional[CompileOptions] = None):
  options = options or CompileOptions()
  if options.mode == 'jax':
    from torchax import interop
    if isinstance(fn, torch.nn.Module):
      module = interop.JittableModule(fn, extra_jit_args=options.jax_jit_kwargs)
      for n in options.methods_to_compile:
        module.make_jitted(n)
      return module
    else:
      return interop.jax_jit(fn)
  elif options.mode == 'dynamo':
    raise RuntimeError('dynamo mode is not supported yet')
  elif options.mode == 'export':
    raise RuntimeError('export mode is not supported yet')


@contextmanager
def jax_device(target_device: str, env: tensor.Environment | None = None):
  """
  to("jax") cannot differentiate the device/platform (cpu vs tpu). 
  Use this context manager to control jax array's storage device
  
  Examples:

  a = torch.ones(3, 3)

  with jax_device("cpu"):
    b = a.to("jax") 
  
  with jax_device("tpu"):
    c = a.to("jax") 
  
  with jax_device("tpu"):
    c = b.to("jax") 

  """
  if env is None:
    env = default_env()

  prev_target_device = env.target_device
  try:
    env.target_device = target_device
    yield env
  finally:
    env.target_device = prev_target_device



import re
import sys
import unittest

import torch
from torch.export import Dim, export
from torchax.export import exported_program_to_stablehlo as exp2shlo
import torchax

## This file is copied from `xla/test/stablehlo/test_unbounded_dynamism.py`
## To test that torchax has identical behavior.
## The only differences in this test files are that torchax export preserves
## argument order more often than torch_xla export.
##
## This broke ~5 tests, for example: test_bmm_dynamic_out_dim
##   args = (
##     torch.rand((8, 128, 256)),
##     torch.rand((8, 256, 3)),
##   )
##   dynamic_shapes = ((None, {2: Dim("dim")}),)
##   ...
##   torch_xla_regex = r'%arg.: tensor<8x256x\?xf32>.*%arg.: tensor<8x128x256xf32>.*->.*tensor<8x128x\?xf32>'
##   torchax_regex = r'%arg.: tensor<8x128x256xf32>.*%arg.: tensor<8x256x\?xf32>.*->.*tensor<8x128x\?xf32>'


# Shim to run tests
class ExportAdapter():

  def __init__(self, export):
    self.export = export

  def get_stablehlo_text(self):
    return self.export.mlir_module()


def exported_program_to_stablehlo(exported):
  return ExportAdapter(exp2shlo(exported)[1])


def wrap_func_as_nn_module(f):

  class M(torch.nn.Module):

    def __init__(self):
      super().__init__()

    def forward(self, *args):
      return f(*args)

  return M().eval()


class UnboundedDynamismExportTest(unittest.TestCase):

  def setUp(self):
    self.env = torchax.default_env()
    self.env.config.use_torch_native_for_cpu_tensor = False
    torchax.enable_accuracy_mode()

  def tearDown(self):
    self.env.config.use_torch_native_for_cpu_tensor = True

  def test_add(self):
    args = (torch.rand((10, 197, 768)), torch.rand((10, 197, 768)))
    dynamic_shapes = (({0: Dim("dim")}, {0: Dim("dim")}),)
    m = wrap_func_as_nn_module(torch.ops.aten.add.Tensor)
    ep = export(m, args=args, dynamic_shapes=dynamic_shapes)
    shlo_module = exported_program_to_stablehlo(ep)
    shlo_text = shlo_module.get_stablehlo_text()
    self.assertTrue(
        re.search(
            r'tensor<\?x197x768xf32>.*tensor<\?x197x768xf32>.*->.*tensor<\?x197x768xf32>',
            shlo_text) is not None)

  def test_add_scalar(self):
    args = (torch.rand((10, 197, 768)), 0.345)
    dynamic_shapes = (({0: Dim("dim")}, None),)
    m = wrap_func_as_nn_module(torch.ops.aten.add.Tensor)
    ep = export(m, args=args, dynamic_shapes=dynamic_shapes)
    shlo_module = exported_program_to_stablehlo(ep)
    shlo_text = shlo_module.get_stablehlo_text()
    self.assertTrue(
        re.search(r'tensor<\?x197x768xf32>.*->.*tensor<\?x197x768xf32>',
                  shlo_text) is not None)

  def test_addmm(self):
    args = (torch.rand((5)), torch.rand((10, 5)), torch.rand((5, 5)))
    dynamic_shapes = ((None, {0: Dim("dim")}, None),)
    m = wrap_func_as_nn_module(torch.ops.aten.addmm.default)
    ep = export(m, args=args, dynamic_shapes=dynamic_shapes)
    shlo_module = exported_program_to_stablehlo(ep)
    shlo_text = shlo_module.get_stablehlo_text()
    self.assertTrue(
        re.search(r'tensor<\?x5xf32>.*->.*tensor<\?x5xf32>', shlo_text)
        is not None)

  def test_bmm(self):
    args = (
        torch.rand((24, 197, 64)),
        torch.rand((24, 64, 197)),
    )
    dynamic_shapes = (({0: Dim("dim")}, {0: Dim("dim")}),)
    m = wrap_func_as_nn_module(torch.ops.aten.bmm.default)
    ep = export(m, args=args, dynamic_shapes=dynamic_shapes)
    shlo_module = exported_program_to_stablehlo(ep)
    shlo_text = shlo_module.get_stablehlo_text()
    self.assertTrue(
        re.search(
            r'%arg.: tensor<\?x197x64xf32>.*%arg.: tensor<\?x64x197xf32>.*->.*tensor<\?x197x197xf32>',
            shlo_text) is not None)

  def test_bmm_dynamic_out_dim(self):
    args = (
        torch.rand((8, 128, 256)),
        torch.rand((8, 256, 3)),
    )
    dynamic_shapes = ((None, {2: Dim("dim")}),)
    m = wrap_func_as_nn_module(torch.ops.aten.bmm.default)
    ep = export(m, args=args, dynamic_shapes=dynamic_shapes)
    shlo_module = exported_program_to_stablehlo(ep)
    shlo_text = shlo_module.get_stablehlo_text()
    self.assertTrue(
        re.search(
            r'%arg.: tensor<8x128x256xf32>.*%arg.: tensor<8x256x\?xf32>.*->.*tensor<8x128x\?xf32>',
            shlo_text) is not None)

  def test_bmm_dynamic_reduction_dim(self):
    args = (
        torch.rand((8, 128, 3)),
        torch.rand((8, 3, 256)),
    )
    dynamic_shapes = (({2: Dim("dim")}, {1: Dim("dim")}),)
    m = wrap_func_as_nn_module(torch.ops.aten.bmm.default)
    ep = export(m, args=args, dynamic_shapes=dynamic_shapes)
    shlo_module = exported_program_to_stablehlo(ep)
    shlo_text = shlo_module.get_stablehlo_text()
    self.assertTrue(
        re.search(
            r'%arg.: tensor<8x128x\?xf32>.*%arg.: tensor<8x\?x256xf32>.*->.*tensor<8x128x256xf32>',
            shlo_text) is not None)

  def test_cat(self):
    args = (torch.rand((10, 1, 768)), torch.rand((10, 196, 768)))
    dynamic_shapes = (({0: Dim("dim")}, {0: Dim("dim")}),)
    m = wrap_func_as_nn_module(
        lambda x, y: torch.ops.aten.cat.default([x, y], 1))
    ep = export(m, args=args, dynamic_shapes=dynamic_shapes)
    shlo_module = exported_program_to_stablehlo(ep)
    shlo_text = shlo_module.get_stablehlo_text()
    self.assertTrue(
        re.search(
            r'%arg.: tensor<\?x1x768xf32>.*%arg.: tensor<\?x196x768xf32>.*->.*tensor<\?x197x768xf32>',
            shlo_text) is not None)

  def test_conv(self):
    args = (
        torch.rand((10, 3, 224, 224)),
        torch.rand((5, 3, 16, 16)),
        torch.rand((5)),
    )
    dynamic_shapes = (({0: Dim("dim")}, None, None),)
    m = wrap_func_as_nn_module(
        lambda x, y, z: torch.ops.aten.convolution.default(
            x,
            y,
            z,
            [16, 16],
            [0, 0],
            [1, 1],
            False,
            [0, 0],
            1,
        ))
    ep = export(m, args=args, dynamic_shapes=dynamic_shapes)
    shlo_module = exported_program_to_stablehlo(ep)
    shlo_text = shlo_module.get_stablehlo_text()
    self.assertTrue(
        re.search(r'tensor<\?x3x224x224xf32>.*->.*tensor<\?x5x14x14xf32>',
                  shlo_text) is not None)

  def test_conv1d(self):
    args = (
        torch.rand((3, 1, 800)),
        torch.rand((512, 1, 10)),
    )
    dynamic_shapes = (({0: Dim("dim")}, None),)
    # dynamic_shapes = None
    m = wrap_func_as_nn_module(lambda x, y: torch.ops.aten.convolution.default(
        x,
        y,
        None,
        [5],
        [0],
        [1],
        False,
        [0],
        1,
    ))
    ep = export(m, args=args, dynamic_shapes=dynamic_shapes)
    shlo_module = exported_program_to_stablehlo(ep)
    shlo_text = shlo_module.get_stablehlo_text()
    self.assertTrue(
        re.search(r'tensor<\?x1x800xf32>.*->.*tensor<\?x512x159xf32>',
                  shlo_text) is not None)

  def test_cumsum(self):
    args = (torch.rand((10, 5)), 1)
    dynamic_shapes = (({0: Dim("dim")}, None),)
    m = wrap_func_as_nn_module(torch.ops.aten.cumsum.default)
    ep = export(m, args=args, dynamic_shapes=dynamic_shapes)
    shlo_module = exported_program_to_stablehlo(ep)
    shlo_text = shlo_module.get_stablehlo_text()
    self.assertTrue(
        re.search(r'tensor<\?x5xf32>.*->.*tensor<\?x5xf32>', shlo_text)
        is not None)

  def test_div(self):
    args = (torch.rand((10, 12, 197)), torch.rand((10, 12, 197)))
    dynamic_shapes = (({0: Dim("dim")}, {0: Dim("dim")}),)
    m = wrap_func_as_nn_module(torch.ops.aten.div.Tensor)
    ep = export(m, args=args, dynamic_shapes=dynamic_shapes)
    shlo_module = exported_program_to_stablehlo(ep)
    shlo_text = shlo_module.get_stablehlo_text()
    self.assertTrue(
        re.search(
            r'tensor<\?x12x197xf32>.*tensor<\?x12x197xf32>.*->.*tensor<\?x12x197xf32>',
            shlo_text) is not None)

  def test_div_scalar(self):
    args = (torch.rand((10, 12, 197)), 8.0)
    dynamic_shapes = (({0: Dim("dim")}, None),)
    m = wrap_func_as_nn_module(torch.ops.aten.div.Tensor)
    ep = export(m, args=args, dynamic_shapes=dynamic_shapes)
    shlo_module = exported_program_to_stablehlo(ep)
    shlo_text = shlo_module.get_stablehlo_text()
    self.assertTrue(
        re.search(r'tensor<\?x12x197xf32>.*->.*tensor<\?x12x197xf32>',
                  shlo_text) is not None)

  def test_gelu(self):
    args = (torch.rand((3, 5)),)
    dynamic_shapes = (({0: Dim("dim")},),)
    m = wrap_func_as_nn_module(torch.ops.aten.gelu)
    ep = export(m, args=args, dynamic_shapes=dynamic_shapes)
    shlo_module = exported_program_to_stablehlo(ep)
    shlo_text = shlo_module.get_stablehlo_text()
    self.assertTrue(
        re.search(r'tensor<\?x5xf32>.*->.*tensor<\?x5xf32>', shlo_text)
        is not None)

  def test_embedding(self):

    class M(torch.nn.Module):

      def forward(self, x, y):
        res = torch.ops.aten.embedding.default(x, y)
        return res

    args = (torch.rand((20, 768)), torch.randint(0, 15,
                                                 (3, 10)).to(torch.int64))
    dynamic_shapes = (None, {0: Dim("dim")})
    m = M()
    ep = export(m, args=args, dynamic_shapes=dynamic_shapes)
    shlo_module = exported_program_to_stablehlo(ep)
    shlo_text = shlo_module.get_stablehlo_text()
    self.assertTrue(
        re.search(r"%arg.: tensor<\?x10xi64>.*->.*tensor<\?x10x768xf32>",
                  shlo_text) is not None)

  def test_mean(self):

    class M(torch.nn.Module):

      def forward(self, x):
        return torch.mean(x, -1, keepdim=True)

    args = (torch.rand((10, 197, 768)),)
    dynamic_shapes = ({0: Dim("dim")},)
    m = M()
    ep = export(m, args=args, dynamic_shapes=dynamic_shapes)
    shlo_module = exported_program_to_stablehlo(ep)
    shlo_text = shlo_module.get_stablehlo_text()
    self.assertTrue(
        re.search(r"%arg.: tensor<\?x197x768xf32>.*->.*tensor<\?x197x1xf32>",
                  shlo_text) is not None)

  def test_mul(self):
    args = (torch.rand((10, 2, 768)), torch.rand((10, 2, 768)))
    dynamic_shapes = (({0: Dim("dim")}, {0: Dim("dim")}),)
    m = wrap_func_as_nn_module(torch.ops.aten.mul.Tensor)
    ep = export(m, args=args, dynamic_shapes=dynamic_shapes)
    shlo_module = exported_program_to_stablehlo(ep)
    shlo_text = shlo_module.get_stablehlo_text()
    self.assertTrue(
        re.search(
            r'tensor<\?x2x768xf32>.*tensor<\?x2x768xf32>.*->.*tensor<\?x2x768xf32>',
            shlo_text) is not None)

  def test_mul_scalar(self):
    args = (torch.rand((10, 2, 768)), 0.125)
    dynamic_shapes = (({0: Dim("dim")}, None),)
    m = wrap_func_as_nn_module(torch.ops.aten.mul.Tensor)
    ep = export(m, args=args, dynamic_shapes=dynamic_shapes)
    shlo_module = exported_program_to_stablehlo(ep)
    shlo_text = shlo_module.get_stablehlo_text()
    self.assertTrue(
        re.search(r'tensor<\?x2x768xf32>.*->.*tensor<\?x2x768xf32>', shlo_text)
        is not None)

  def test_ne_scalar(self):

    class M(torch.nn.Module):

      def forward(self, x):
        return torch.ops.aten.ne.Scalar(x, 1).to(torch.int32)

    args = (torch.rand((3, 5)).to(torch.int64),)
    dynamic_shapes = ({0: Dim("dim")},)
    m = M()
    ep = export(m, args=args, dynamic_shapes=dynamic_shapes)
    shlo_module = exported_program_to_stablehlo(ep)
    shlo_text = shlo_module.get_stablehlo_text()
    self.assertTrue(
        re.search(r"%arg.: tensor<\?x5xi64>.*->.*tensor<\?x5xi32>", shlo_text)
        is not None)

  def test_var(self):

    class M(torch.nn.Module):

      def forward(self, x):
        return torch.var(x, -1, keepdim=True, correction=0)

    args = (torch.rand((10, 197, 768)),)
    dynamic_shapes = ({0: Dim("dim")},)
    m = M()
    ep = export(m, args=args, dynamic_shapes=dynamic_shapes)
    shlo_module = exported_program_to_stablehlo(ep)
    shlo_text = shlo_module.get_stablehlo_text()
    self.assertTrue(
        re.search(r"%arg.: tensor<\?x197x768xf32>.*->.*tensor<\?x197x1xf32>",
                  shlo_text) is not None)

  def test_native_group_norm(self):

    class M2(torch.nn.Module):

      def __init__(self):
        super().__init__()
        self.layer_norm = torch.nn.GroupNorm(
            num_groups=512, num_channels=512, affine=True)

      def forward(self, x):
        x = self.layer_norm(x)
        return x

    args = (torch.rand((10, 512, 159)),)
    dynamic_shapes = ({0: Dim("dim")},)
    m = M2()
    out1 = m(*args)
    ep = export(m, args=args, dynamic_shapes=dynamic_shapes)
    shlo_module = exported_program_to_stablehlo(ep)
    shlo_text = shlo_module.get_stablehlo_text()
    self.assertTrue(
        re.search(r"%arg.: tensor<\?x512x159xf32>.*->.*tensor<\?x512x159xf32>",
                  shlo_text) is not None)

  def test_native_layer_norm(self):

    class M(torch.nn.Module):

      def forward(self, x, weight, bias):
        return torch.ops.aten.native_layer_norm.default(x, [768], weight, bias,
                                                        1e-12)[0]

    args = (
        torch.rand((10, 197, 768)),
        torch.rand((768)),
        torch.rand((768)),
    )
    dynamic_shapes = ({0: Dim("dim")}, None, None)
    m = M()
    ep = export(m, args=args, dynamic_shapes=dynamic_shapes)
    shlo_module = exported_program_to_stablehlo(ep)
    shlo_text = shlo_module.get_stablehlo_text()
    self.assertTrue(
        re.search(r"%arg.: tensor<\?x197x768xf32>.*->.*tensor<\?x197x768xf32>",
                  shlo_text) is not None)

  def test_permute(self):
    args = (torch.rand((10, 197, 12, 64)),)
    dynamic_shapes = (({0: Dim("dim")},),)
    m = wrap_func_as_nn_module(
        lambda x: torch.ops.aten.permute.default(x, [0, 2, 1, 3]))
    ep = export(m, args=args, dynamic_shapes=dynamic_shapes)
    shlo_module = exported_program_to_stablehlo(ep)
    shlo_text = shlo_module.get_stablehlo_text()
    self.assertTrue(
        re.search(
            r"%arg.: tensor<\?x197x12x64xf32>.*->.*tensor<\?x12x197x64xf32>",
            shlo_text) is not None)

  def test_select(self):
    args = (torch.rand((10, 197, 768)), 1, 0)
    dynamic_shapes = (({0: Dim("dim")}, None, None),)
    m = wrap_func_as_nn_module(torch.ops.aten.select.int)
    ep = export(m, args=args, dynamic_shapes=dynamic_shapes)
    shlo_module = exported_program_to_stablehlo(ep)
    shlo_text = shlo_module.get_stablehlo_text()
    self.assertTrue(
        re.search(r"%arg.: tensor<\?x197x768xf32>.*->.*tensor<\?x768xf32>",
                  shlo_text) is not None)

  def test_slice(self):
    args = (torch.rand((10, 3, 224, 224)), 0, 0, 9223372036854775807)
    dynamic_shapes = (({0: Dim("dim")}, None, None, None),)
    m = wrap_func_as_nn_module(torch.ops.aten.slice.Tensor)
    ep = export(m, args=args, dynamic_shapes=dynamic_shapes)
    shlo_module = exported_program_to_stablehlo(ep)
    shlo_text = shlo_module.get_stablehlo_text()
    self.assertTrue(
        re.search(
            r"%arg.: tensor<\?x3x224x224xf32>.*->.*tensor<\?x3x224x224xf32>",
            shlo_text) is not None)

  def test_slice_2(self):
    args = (torch.rand((10, 3, 224, 224)), 1, 0, 2)
    dynamic_shapes = (({0: Dim("dim")}, None, None, None),)
    m = wrap_func_as_nn_module(torch.ops.aten.slice.Tensor)
    ep = export(m, args=args, dynamic_shapes=dynamic_shapes)
    shlo_module = exported_program_to_stablehlo(ep)
    shlo_text = shlo_module.get_stablehlo_text()
    self.assertTrue(
        re.search(
            r"%arg.: tensor<\?x3x224x224xf32>.*->.*tensor<\?x2x224x224xf32>",
            shlo_text) is not None)

  def test_softmax(self):
    args = (torch.rand((10, 12, 197, 197)), -1, False)
    dynamic_shapes = (({0: Dim("dim")}, None, None),)
    m = wrap_func_as_nn_module(torch.ops.aten._softmax.default)
    ep = export(m, args=args, dynamic_shapes=dynamic_shapes)
    shlo_module = exported_program_to_stablehlo(ep)
    shlo_text = shlo_module.get_stablehlo_text()
    self.assertTrue(
        re.search(
            r"%arg.: tensor<\?x12x197x197xf32>.*->.*tensor<\?x12x197x197xf32>",
            shlo_text) is not None)

  def test_sub(self):
    args = (torch.rand((10, 1, 1, 10)), torch.rand((10, 1, 1, 10)))
    dynamic_shapes = (({0: Dim("dim")}, {0: Dim("dim")}),)
    m = wrap_func_as_nn_module(torch.ops.aten.sub.Tensor)
    ep = export(m, args=args, dynamic_shapes=dynamic_shapes)
    shlo_module = exported_program_to_stablehlo(ep)
    shlo_text = shlo_module.get_stablehlo_text()
    self.assertTrue(
        re.search(
            r'tensor<\?x1x1x10xf32>.*tensor<\?x1x1x10xf32>.*->.*tensor<\?x1x1x10xf32>',
            shlo_text) is not None)

  def test_softmax_reduce_on_dynamic_dim(self):
    args = (torch.rand((1, 8, 128, 3)), -1, False)
    dynamic_shapes = (({3: Dim("dim")}, None, None),)
    m = wrap_func_as_nn_module(torch.ops.aten._softmax.default)
    ep = export(m, args=args, dynamic_shapes=dynamic_shapes)
    shlo_module = exported_program_to_stablehlo(ep)
    shlo_text = shlo_module.get_stablehlo_text()
    self.assertTrue(
        re.search(r"%arg.: tensor<1x8x128x\?xf32>.*->.*tensor<1x8x128x\?xf32>",
                  shlo_text) is not None)

  @unittest.skip("Converted StableHLO contains i1 dtype, not expected.")
  def test_index(self):
    args = (torch.rand((2, 10)), torch.arange(5))
    dynamic_shapes = ((None, {0: Dim("dim")}),)
    m = wrap_func_as_nn_module(
        lambda x, y: torch.ops.aten.index.Tensor(x, [None, y]))
    ep = export(m, args=args, dynamic_shapes=dynamic_shapes)
    shlo_module = exported_program_to_stablehlo(ep)
    shlo_text = shlo_module.get_stablehlo_text()
    self.assertTrue(
        re.search(
            r"%arg.: tensor<\?xi64>.*%arg.: tensor<2x10xf32>.*->.*tensor<2x\?xf32>",
            shlo_text) is not None)

  def test_sub_scalar(self):
    args = (1.0, torch.rand((10, 1, 1, 10)))
    dynamic_shapes = ((None, {0: Dim("dim")}),)
    m = wrap_func_as_nn_module(torch.ops.aten.sub.Tensor)
    ep = export(m, args=args, dynamic_shapes=dynamic_shapes)
    shlo_module = exported_program_to_stablehlo(ep)
    shlo_text = shlo_module.get_stablehlo_text()
    self.assertTrue(
        re.search(r'tensor<\?x1x1x10xf32>.*->.*tensor<\?x1x1x10xf32>',
                  shlo_text) is not None)

  def test_split_with_sizes(self):

    class M(torch.nn.Module):

      def forward(self, x):
        res = torch.ops.aten.split_with_sizes.default(x, [1, 2, 3], -1)
        return res[0], res[1], res[2]

    args = (torch.rand((3, 10, 6)),)
    dynamic_shapes = ({0: Dim("dim")},)
    m = M()
    ep = export(m, args=args, dynamic_shapes=dynamic_shapes)
    shlo_module = exported_program_to_stablehlo(ep)
    shlo_text = shlo_module.get_stablehlo_text()
    self.assertTrue(
        re.search(
            r"%arg.: tensor<\?x10x6xf32>.*->.*tensor<\?x10x1xf32>.*tensor<\?x10x2xf32>.*tensor<\?x10x3xf32>",
            shlo_text) is not None)

  def test_transpose_on_dynamic_dim(self):
    args = (torch.rand((1, 8, 3, 256)),)
    dynamic_shapes = (({2: Dim("dim")},),)
    m = wrap_func_as_nn_module(
        lambda x: torch.ops.aten.transpose.int(x, -2, -1))
    ep = export(m, args=args, dynamic_shapes=dynamic_shapes)
    shlo_module = exported_program_to_stablehlo(ep)
    shlo_text = shlo_module.get_stablehlo_text()
    self.assertTrue(
        re.search(r"%arg.: tensor<1x8x\?x256xf32>.*->.*tensor<1x8x256x\?xf32>",
                  shlo_text) is not None)

  def test_unsqueeze_1(self):
    args = (torch.rand((3, 10)),)
    dynamic_shapes = (({0: Dim("dim")},),)
    m = wrap_func_as_nn_module(lambda x: torch.ops.aten.unsqueeze.default(x, 1))
    ep = export(m, args=args, dynamic_shapes=dynamic_shapes)
    shlo_module = exported_program_to_stablehlo(ep)
    shlo_text = shlo_module.get_stablehlo_text()
    self.assertTrue(
        re.search(r"%arg.: tensor<\?x10xf32>.*->.*tensor<\?x1x10xf32>",
                  shlo_text) is not None)

  def test_unsqueeze_2(self):
    args = (torch.rand((1, 1, 3, 256)),)
    dynamic_shapes = (({2: Dim("dim")},),)
    m = wrap_func_as_nn_module(lambda x: torch.ops.aten.unsqueeze.default(x, 2))
    ep = export(m, args=args, dynamic_shapes=dynamic_shapes)
    shlo_module = exported_program_to_stablehlo(ep)
    shlo_text = shlo_module.get_stablehlo_text()
    self.assertTrue(
        re.search(
            r"%arg.: tensor<1x1x\?x256xf32>.*->.*tensor<1x1x1x\?x256xf32>",
            shlo_text) is not None)

  def test_dynamic_view(self):

    class M(torch.nn.Module):

      def __init__(self):
        super().__init__()
        self.conv = torch.nn.Conv2d(3, 5, [16, 16])

      def forward(self, x):
        x = self.conv(x)
        return x.view(x.shape[0], x.shape[1], -1)

    m = M().eval()
    args = (torch.rand((10, 3, 224, 224)),)
    dynamic_shapes = ({0: Dim("bs")},)
    ep = export(m, args, dynamic_shapes=dynamic_shapes)
    out1 = ep.module()(*args)
    shlo_module = exported_program_to_stablehlo(ep)
    shlo_text = shlo_module.get_stablehlo_text()
    self.assertTrue(
        re.search(
            r"%arg.: tensor<\?x3x224x224xf32>.*->.*tensor<\?x5x43681xf32>",
            shlo_text) is not None)

  @unittest.skip("Cannot generate aten.sym_numel in the exported program.")
  def test_dynamic_view_sym_numel(self):

    class M(torch.nn.Module):

      def forward(self, x, range):
        num_elem = torch.numel(range)
        return x.view(x.shape[0], x.shape[2], num_elem, x.shape[4])

    m = M().eval()
    args = (torch.rand((1, 1, 8, 3, 256)), torch.arange(3))
    dynamic_shapes = ({3: Dim("bs")}, {0: Dim("bs")})
    ep = export(m, args, dynamic_shapes=dynamic_shapes)
    out1 = ep.module()(*args)
    shlo_module = exported_program_to_stablehlo(ep)
    shlo_text = shlo_module.get_stablehlo_text()
    self.assertTrue(
        re.search(
            r"%arg.: tensor<\?x3x224x224xf32>.*->.*tensor<\?x5x43681xf32>",
            shlo_text) is not None)

  def test_dynamic_view_non_bs(self):

    class M(torch.nn.Module):

      def forward(self, x):
        return x.view(x.shape[0], x.shape[1] * x.shape[2], x.shape[3])

    m = M().eval()
    args = (torch.rand((1, 3, 2, 16)),)
    dynamic_shapes = ({1: Dim("bs")},)
    ep = export(m, args, dynamic_shapes=dynamic_shapes)
    out1 = ep.module()(*args)
    shlo_module = exported_program_to_stablehlo(ep)
    shlo_text = shlo_module.get_stablehlo_text()
    self.assertTrue(
        re.search(r"%arg.: tensor<1x\?x2x16xf32>.*->.*tensor<1x\?x16xf32>",
                  shlo_text) is not None)

  def test_dynamic_view_multiplier(self):

    class M(torch.nn.Module):

      def forward(self, x):
        return x.view(x.shape[0] * x.shape[1], -1)

    m = M().eval()
    args = (torch.rand((10, 197, 768)),)
    dynamic_shapes = ({0: Dim("bs")},)
    ep = export(m, args, dynamic_shapes=dynamic_shapes)
    out1 = ep.module()(*args)
    shlo_module = exported_program_to_stablehlo(ep)
    shlo_text = shlo_module.get_stablehlo_text()
    self.assertTrue(
        re.search(r"%arg.: tensor<\?x197x768xf32>.*->.*tensor<\?x768xf32>",
                  shlo_text) is not None)

  def test_dynamic_expand(self):

    class M(torch.nn.Module):

      def forward(self, x, image):
        return x.expand(image.shape[0], -1, -1)

    m = M().eval()
    args = (torch.rand((1, 1, 768)), torch.rand((10, 3, 224, 224)))
    dynamic_shapes = (
        None,
        {
            0: Dim("bs")
        },
    )
    ep = export(m, args, dynamic_shapes=dynamic_shapes)
    out1 = ep.module()(*args)
    shlo_module = exported_program_to_stablehlo(ep)
    shlo_text = shlo_module.get_stablehlo_text()
    self.assertTrue(
        re.search(r"%arg.: tensor<1x1x768xf32>.*->.*tensor<\?x1x768xf32>",
                  shlo_text) is not None)

  def test_dynamic_expand_2(self):

    class M(torch.nn.Module):

      def forward(self, x, range):
        return x.expand(1, 1, 8, range.shape[0], 256)

    m = M().eval()
    args = (torch.rand((1, 1, 1, 3, 256)), torch.arange(3))
    dynamic_shapes = ({3: Dim("bs")}, {0: Dim("bs")})
    ep = export(m, args, dynamic_shapes=dynamic_shapes)
    out1 = ep.module()(*args)
    shlo_module = exported_program_to_stablehlo(ep)
    shlo_text = shlo_module.get_stablehlo_text()
    self.assertTrue(
        re.search(
            r"%arg.: tensor<1x1x1x\?x256xf32>.*->.*tensor<1x1x8x\?x256xf32>",
            shlo_text) is not None)


if __name__ == "__main__":
  test = unittest.main()
  sys.exit(0 if test.result.wasSuccessful() else 1)



"""
Forked at: https://raw.githubusercontent.com/mlperf/training_results_v0.7/refs/heads/master/Google/benchmarks/ssd/implementations/ssd-research-JAX-tpu-v3-4096/nms.py
"""

import functools
from typing import List, Union, Optional, Tuple

import torch
from jax import lax
import jax.numpy as jnp
from . import ops_registry

_NMS_TILE_SIZE = 256


def _bbox_overlap(boxes, gt_boxes):
  """Find Bounding box overlap.

  Args:
    boxes: first set of bounding boxes
    gt_boxes: second set of boxes to compute IOU

  Returns:
    iou: Intersection over union matrix of all input bounding boxes
  """
  bb_y_min, bb_x_min, bb_y_max, bb_x_max = jnp.split(
      ary=boxes, indices_or_sections=4, axis=2)
  gt_y_min, gt_x_min, gt_y_max, gt_x_max = jnp.split(
      ary=gt_boxes, indices_or_sections=4, axis=2)

  # Calculates the intersection area.
  i_xmin = jnp.maximum(bb_x_min, jnp.transpose(gt_x_min, [0, 2, 1]))
  i_xmax = jnp.minimum(bb_x_max, jnp.transpose(gt_x_max, [0, 2, 1]))
  i_ymin = jnp.maximum(bb_y_min, jnp.transpose(gt_y_min, [0, 2, 1]))
  i_ymax = jnp.minimum(bb_y_max, jnp.transpose(gt_y_max, [0, 2, 1]))
  i_area = jnp.maximum((i_xmax - i_xmin), 0) * jnp.maximum((i_ymax - i_ymin), 0)

  # Calculates the union area.
  bb_area = (bb_y_max - bb_y_min) * (bb_x_max - bb_x_min)
  gt_area = (gt_y_max - gt_y_min) * (gt_x_max - gt_x_min)
  # Adds a small epsilon to avoid divide-by-zero.
  u_area = bb_area + jnp.transpose(gt_area, [0, 2, 1]) - i_area + 1e-8

  # Calculates IoU.
  iou = i_area / u_area

  return iou


def _self_suppression(in_args):
  iou, _, iou_sum = in_args
  batch_size = iou.shape[0]
  can_suppress_others = jnp.reshape(
      jnp.max(iou, 1) <= 0.5, [batch_size, -1, 1]).astype(iou.dtype)
  iou_suppressed = jnp.reshape(
      (jnp.max(can_suppress_others * iou, 1) <= 0.5).astype(
          iou.dtype), [batch_size, -1, 1]) * iou
  iou_sum_new = jnp.sum(iou_suppressed, [1, 2])
  return iou_suppressed, jnp.any(iou_sum - iou_sum_new > 0.5), iou_sum_new


def _cross_suppression(in_args):
  boxes, box_slice, iou_threshold, inner_idx = in_args
  batch_size = boxes.shape[0]
  new_slice = lax.dynamic_slice(boxes, [0, inner_idx * _NMS_TILE_SIZE, 0],
                                [batch_size, _NMS_TILE_SIZE, 4])
  iou = _bbox_overlap(new_slice, box_slice)
  ret_slice = jnp.expand_dims((jnp.all(iou < iou_threshold, [1])).astype(
      box_slice.dtype), 2) * box_slice
  return boxes, ret_slice, iou_threshold, inner_idx + 1


def _suppression_loop_body(in_args):
  """Process boxes in the range [idx*_NMS_TILE_SIZE, (idx+1)*_NMS_TILE_SIZE).

  Args:
    in_args: A tuple of arguments: boxes, iou_threshold, output_size, idx

  Returns:
    boxes: updated boxes.
    iou_threshold: pass down iou_threshold to the next iteration.
    output_size: the updated output_size.
    idx: the updated induction variable.
  """
  boxes, iou_threshold, output_size, idx = in_args
  num_tiles = boxes.shape[1] // _NMS_TILE_SIZE
  batch_size = boxes.shape[0]

  # Iterates over tiles that can possibly suppress the current tile.
  box_slice = lax.dynamic_slice(boxes, [0, idx * _NMS_TILE_SIZE, 0],
                                [batch_size, _NMS_TILE_SIZE, 4])

  def _loop_cond(in_args):
    _, _, _, inner_idx = in_args
    return inner_idx < idx

  _, box_slice, _, _ = lax.while_loop(_loop_cond, _cross_suppression,
                                      (boxes, box_slice, iou_threshold, 0))

  # Iterates over the current tile to compute self-suppression.
  iou = _bbox_overlap(box_slice, box_slice)
  mask = jnp.expand_dims(
      jnp.reshape(jnp.arange(_NMS_TILE_SIZE), [1, -1])
      > jnp.reshape(jnp.arange(_NMS_TILE_SIZE), [-1, 1]), 0)
  iou *= (jnp.logical_and(mask, iou >= iou_threshold)).astype(iou.dtype)

  def _loop_cond2(in_args):
    _, loop_condition, _ = in_args
    return loop_condition

  suppressed_iou, _, _ = lax.while_loop(_loop_cond2, _self_suppression,
                                        (iou, True, jnp.sum(iou, [1, 2])))
  suppressed_box = jnp.sum(suppressed_iou, 1) > 0
  box_slice *= jnp.expand_dims(1.0 - suppressed_box.astype(box_slice.dtype), 2)

  # Uses box_slice to update the input boxes.
  mask = jnp.reshape((jnp.equal(jnp.arange(num_tiles),
                                idx)).astype(boxes.dtype), [1, -1, 1, 1])
  boxes = jnp.tile(jnp.expand_dims(
      box_slice, 1), [1, num_tiles, 1, 1]) * mask + jnp.reshape(
          boxes, [batch_size, num_tiles, _NMS_TILE_SIZE, 4]) * (1 - mask)
  boxes = jnp.reshape(boxes, [batch_size, -1, 4])

  # Updates output_size.
  output_size += jnp.sum(jnp.any(box_slice > 0, [2]).astype(jnp.int32), [1])
  return boxes, iou_threshold, output_size, idx + 1


def non_max_suppression_padded(scores, boxes, max_output_size, iou_threshold):
  """A wrapper that handles non-maximum suppression.

  Assumption:
    * The boxes are sorted by scores unless the box is a dot (all coordinates
      are zero).
    * Boxes with higher scores can be used to suppress boxes with lower scores.

  The overal design of the algorithm is to handle boxes tile-by-tile:

  boxes = boxes.pad_to_multiply_of(tile_size)
  num_tiles = len(boxes) // tile_size
  output_boxes = []
  for i in range(num_tiles):
    box_tile = boxes[i*tile_size : (i+1)*tile_size]
    for j in range(i - 1):
      suppressing_tile = boxes[j*tile_size : (j+1)*tile_size]
      iou = _bbox_overlap(box_tile, suppressing_tile)
      # if the box is suppressed in iou, clear it to a dot
      box_tile *= _update_boxes(iou)
    # Iteratively handle the diagnal tile.
    iou = _box_overlap(box_tile, box_tile)
    iou_changed = True
    while iou_changed:
      # boxes that are not suppressed by anything else
      suppressing_boxes = _get_suppressing_boxes(iou)
      # boxes that are suppressed by suppressing_boxes
      suppressed_boxes = _get_suppressed_boxes(iou, suppressing_boxes)
      # clear iou to 0 for boxes that are suppressed, as they cannot be used
      # to suppress other boxes any more
      new_iou = _clear_iou(iou, suppressed_boxes)
      iou_changed = (new_iou != iou)
      iou = new_iou
    # remaining boxes that can still suppress others, are selected boxes.
    output_boxes.append(_get_suppressing_boxes(iou))
    if len(output_boxes) >= max_output_size:
      break

  Args:
    scores: a tensor with a shape of [batch_size, anchors].
    boxes: a tensor with a shape of [batch_size, anchors, 4].
    max_output_size: a scalar integer `Tensor` representing the maximum number
      of boxes to be selected by non max suppression.
    iou_threshold: a float representing the threshold for deciding whether boxes
      overlap too much with respect to IOU.
  Returns:
    nms_scores: a tensor with a shape of [batch_size, anchors]. It has same
      dtype as input scores.
    nms_proposals: a tensor with a shape of [batch_size, anchors, 4]. It has
      same dtype as input boxes.
  """
  batch_size = boxes.shape[0]
  num_boxes = boxes.shape[1]
  pad = int(jnp.ceil(
      float(num_boxes) / _NMS_TILE_SIZE)) * _NMS_TILE_SIZE - num_boxes
  boxes = jnp.pad(boxes.astype(jnp.float32), [[0, 0], [0, pad], [0, 0]])
  scores = jnp.pad(scores.astype(jnp.float32), [[0, 0], [0, pad]])
  num_boxes += pad

  def _loop_cond(in_args):
    unused_boxes, unused_threshold, output_size, idx = in_args
    return jnp.logical_and(
        jnp.min(output_size) < max_output_size, idx
        < num_boxes // _NMS_TILE_SIZE)

  selected_boxes, _, output_size, _ = lax.while_loop(
      _loop_cond, _suppression_loop_body,
      (boxes, iou_threshold, jnp.zeros([batch_size], jnp.int32), 0))
  idx = num_boxes - lax.top_k(
      jnp.any(selected_boxes > 0, [2]).astype(jnp.int32) *
      jnp.expand_dims(jnp.arange(num_boxes, 0, -1), 0),
      max_output_size)[0].astype(jnp.int32)
  idx = jnp.minimum(idx, num_boxes - 1)
  idx = jnp.reshape(
      idx + jnp.reshape(jnp.arange(batch_size) * num_boxes, [-1, 1]), [-1])

  return idx
  boxes = jnp.reshape((jnp.reshape(boxes, [-1, 4]))[idx],
                      [batch_size, max_output_size, 4])
  boxes = boxes * (jnp.reshape(jnp.arange(max_output_size), [1, -1, 1])
                   < jnp.reshape(output_size, [-1, 1, 1])).astype(boxes.dtype)
  scores = jnp.reshape(
      jnp.reshape(scores, [-1, 1])[idx], [batch_size, max_output_size])
  scores = scores * (jnp.reshape(jnp.arange(max_output_size), [1, -1])
                     < jnp.reshape(output_size, [-1, 1])).astype(scores.dtype)
  return scores, boxes


# registry:


def nms(boxes, scores, iou_threshold):
  max_output_size = boxes.shape[0]
  boxes = boxes.reshape((1, *boxes.shape))
  scores = scores.reshape((1, *scores.shape))
  res = non_max_suppression_padded(scores, boxes, max_output_size,
                                   iou_threshold)
  return res


try:
  import torch
  import torchvision
  ops_registry.register_torch_dispatch_op(torch.ops.torchvision.nms, nms)
except Exception:
  pass



import collections
import copy
import functools
import torch
from inspect import signature
from functools import wraps
from torch.nn.utils import stateless as torch_stateless
import jax
import jax.numpy as jnp
from jax import tree_util as pytree
from jax.experimental.shard_map import shard_map
from torchax import tensor
from torchax import util
import torchax

from torchax.types import JaxValue, TorchValue, JaxCallable, TorchCallable


def extract_all_buffers(m: torch.nn.Module):
  buffers = {}
  params = {}

  def extract_one(module, prefix):
    for k in dir(module):
      try:
        v = getattr(module, k)
      except:
        continue
      qual_name = prefix + k
      if isinstance(v, torch.nn.parameter.Parameter) and v.requires_grad:
        params[qual_name] = v
      elif isinstance(v, torch.Tensor):
        buffers[qual_name] = v
    for name, child in module.named_children():
      extract_one(child, prefix + name + '.')

  extract_one(m, '')
  return params, buffers


def set_all_buffers(m, params, buffers):

  def set_one(module, prefix):
    for k in dir(module):
      qual_name = prefix + k
      if (potential_v := buffers.get(qual_name)) is not None:
        setattr(module, k, potential_v)
      elif (potential_v := params.get(qual_name)) is not None:
        print(k, potential_v)
        setattr(module, k, torch.nn.Parameter(potential_v))
    for name, child in module.named_children():
      set_one(child, prefix + name + '.')

  set_one(m, '')


class JittableModule(torch.nn.Module):

  def __init__(self,
               m: torch.nn.Module,
               extra_jit_args={},
               dedup_parameters=True):
    super().__init__()
    self.params, self.buffers = extract_all_buffers(m)
    self._model = m
    self._jitted = {}

    self._extra_jit_args = extra_jit_args

    self._extra_dumped_weights = {}

    if dedup_parameters:
      temp = collections.defaultdict(list)
      for k, v in self.params.items():
        temp[id(v)].append(k)

      for v in temp.values():
        if len(v) > 1:
          # duplicated weights with different name
          self._extra_dumped_weights[v[0]] = v[1:]
          for extra_keys in v[1:]:
            del self.params[extra_keys]

  @property
  def __class__(self):
    # Lie about the class type so that
    # isinstance(jittable_module, self._model.__class__) works
    return self._model.__class__

  def __call__(self, *args, **kwargs):
    return self.forward(*args, **kwargs)

  def functional_call(self, method_name, params, buffers, *args, **kwargs):
    kwargs = kwargs or {}
    params_copy = copy.copy(params)
    params_copy.update(buffers)
    # reinflate the state dict so there are not any missing keys
    for k, v in self._extra_dumped_weights.items():
      for new_key in v:
        params_copy[new_key] = params_copy[k]
    with torch_stateless._reparametrize_module(self._model, params_copy):
      res = getattr(self._model, method_name)(*args, **kwargs)
    return res

  def forward(self, *args, **kwargs):
    if 'forward' not in self._jitted:
      jitted = jax_jit(
          functools.partial(self.functional_call, 'forward'),
          kwargs_for_jax_jit=self._extra_jit_args,
      )

      def jitted_forward(*args, **kwargs):
        return jitted(self.params, self.buffers, *args, **kwargs)

      self._jitted['forward'] = jitted_forward
    return self._jitted['forward'](*args, **kwargs)

  def __getattr__(self, key):
    if key == '_model':
      return super().__getattr__(key)
    if key in self._jitted:
      return self._jitted[key]
    return getattr(self._model, key)

  def make_jitted(self, key):
    jitted = jax_jit(
        functools.partial(self.functional_call, key),
        kwargs_for_jax_jit=self._extra_jit_args)

    def call(*args, **kwargs):
      return jitted(self.params, self.buffers, *args, **kwargs)

    self._jitted[key] = call


class CompileMixin:

  def functional_call(self, method, params, buffers, *args, **kwargs):
    kwargs = kwargs or {}
    params_copy = copy.copy(params)
    params_copy.update(buffers)
    with torch_stateless._reparametrize_module(self, params_copy):
      res = method(*args, **kwargs)
    return res

  def jit(self, method):
    jitted = jax_jit(functools.partial(self.functional_call, method_name))

    def call(*args, **kwargs):
      return jitted(self.named_paramters(), self.named_buffers(), *args,
                    **kwargs)

    return call


def compile_nn_module(m: torch.nn.Module, methods=None):
  if methods is None:
    methods = ['forward']

  new_parent = type(
      m.__class__.__name__ + '_with_CompileMixin',
      (CompileMixin, m.__class__),
  )
  m.__class__ = NewParent


def _torch_view(t: JaxValue) -> TorchValue:
  # t is an object from jax land
  # view it as-if it's a torch land object
  if isinstance(t, jax.Array):
    # TODO
    return tensor.Tensor(t, torchax.default_env())
  if isinstance(t, type(jnp.int32)):
    return tensor.t2j_type(t)
  if callable(t):  # t is a JaxCallable
    return functools.partial(call_jax, t)
  # regular types are not changed
  return t


torch_view = functools.partial(pytree.tree_map, _torch_view)


def _jax_view(t: TorchValue) -> JaxValue:
  # t is an object from torch land
  # view it as-if it's a jax land object
  if isinstance(t, torch.Tensor):
    assert isinstance(t, tensor.Tensor) or isinstance(t, tensor.View), type(t)
    return t.jax()
  if isinstance(t, type(torch.int32)):
    return tensor.t2j_dtype(t)

  # torch.nn.Module needs special handling
  if not isinstance(t, torch.nn.Module) and callable(t):  # t is a TorchCallable
    return functools.partial(call_torch, t)
  # regular types are not changed
  return t


jax_view = functools.partial(pytree.tree_map, _jax_view)


def call_jax(jax_func: JaxCallable, *args: TorchValue,
             **kwargs: TorchValue) -> TorchValue:
  args, kwargs = jax_view((args, kwargs))
  res: JaxValue = jax_func(*args, **kwargs)
  return torch_view(res)


def call_torch(torch_func: TorchCallable, *args: JaxValue,
               **kwargs: JaxValue) -> JaxValue:
  args, kwargs = torch_view((args, kwargs))
  with torchax.default_env():
    res: TorchValue = torch_func(*args, **kwargs)
  return jax_view(res)


def j2t_autograd(fn, call_jax=call_jax):
  """Given a JAX function, returns a PyTorch autograd function implemented with `jax.vjp(fn)`.

    It wraps `fn` with `jax.vjp` to compute both the output and residuals (intermediate
    activations). The wrapped function is then run via `call_jax` and integrated into
    the PyTorch autograd framework by saving the residuals into the context object.
    """

  @wraps(fn)
  def inner(*args, **kwargs):
    from jax.tree_util import tree_flatten, tree_unflatten
    from jax.util import safe_zip

    class JaxFun(torch.autograd.Function):

      @staticmethod
      def forward(ctx, tree_def, *flat_args_kwargs):

        tensors, other = util.partition(flat_args_kwargs,
                                        lambda x: isinstance(x, torch.Tensor))
        # We want the arguments that don't require grads to be closured?

        y, fun_vjp = call_jax(_jax_forward, fn, other, tree_def, tensors)

        # Save necessary information for backward
        # Flatten the vjp function. `vjp_spec` contains a jaxpr for the backward pass.
        # `residuals` contains the tensors needed for the backward pass.`
        residuals, vjp_spec = tree_flatten(fun_vjp)
        ctx.vjp_spec = vjp_spec
        ctx.save_for_backward(*residuals)
        return y

      @staticmethod
      def backward(ctx, *grad_out):
        assert len(grad_out) > 0
        grad_out = grad_out if len(grad_out) > 1 else grad_out[0]

        input_grads_structured = call_jax(_jax_backward, ctx.vjp_spec,
                                          ctx.saved_tensors, grad_out)

        # Construct the gradient tuple to be returned.
        # It needs to match the inputs to forward: (tree_def, *flat_inputs)
        # The first gradient (for tree_def) is None.
        # The subsequent gradients correspond to flat_inputs.
        # We need to put a None for inputs that did not require gradients.
        final_grads = [None]
        for needs_grad, grad in safe_zip(ctx.needs_input_grad[1:],
                                         input_grads_structured):
          final_grads.append(grad if needs_grad else None)

        return tuple(final_grads)

    sig = signature(fn)
    bound = sig.bind(*args, **kwargs)
    bound.apply_defaults()
    flat_args_kwargs, tree_def = tree_flatten((bound.args, bound.kwargs))
    y = JaxFun.apply(tree_def, *flat_args_kwargs)
    return y

  return inner


# NOTE(qihqi): This function cannot be inlined from the callsite
#  Becuase if it does, then it won't hit the compilation cache for
#  call_jax. Call jax uses functions' id as key.
def _jax_forward(fn, other, tree_def, tensors):
  """JAX function to compute output and vjp function.

  primals should be a tuple (args, kwargs).
  """
  import jax
  from jax.tree_util import tree_flatten, tree_unflatten

  def fn_wrapper(*tensors):
    # Reconstruct the original args and kwargs
    flat_inputs = util.merge(tensors, other)
    args, kwargs = tree_unflatten(tree_def, flat_inputs)
    return fn(*args, **kwargs)

  return jax.vjp(fn_wrapper, *tensors)


def _jax_backward(vjp_spec, saved_tensors, grad_out):
  """JAX function to compute input gradients.

  Unflattening `saved_tensors` with `vjp_spec` should restore the original vjp function.
  """
  from jax.tree_util import tree_unflatten
  fun_vjp = tree_unflatten(vjp_spec, saved_tensors)
  return fun_vjp(grad_out)


fori_loop = torch_view(jax.lax.fori_loop)


def wrap_jax_jit(torch_function, jax_jit_func=jax.jit, kwargs_for_jax=None):
  kwargs_for_jax = kwargs_for_jax or {}
  jax_func = jax_view(torch_function)
  jitted = jax_jit_func(jax_func, **kwargs_for_jax)
  return torch_view(jitted)


def jax_jit(torch_function,
            kwargs_for_jax_jit=None,
            fix_for_buffer_donation=False):
  return wrap_jax_jit(
      torch_function, jax_jit_func=jax.jit, kwargs_for_jax=kwargs_for_jax_jit)


def jax_shard_map(torch_function, kwargs_for_jax_shard_map=None):
  return wrap_jax_jit(
      torch_function,
      jax_jit_func=shard_map,
      kwargs_for_jax=kwargs_for_jax_shard_map)


def jax_value_and_grad(torch_function, kwargs_for_value_and_grad=None):
  return wrap_jax_jit(
      torch_function,
      jax_jit_func=jax.value_and_grad,
      kwargs_for_jax=kwargs_for_value_and_grad)


def gradient_checkpoint(torch_function, kwargs=None):
  return wrap_jax_jit(
      torch_function, jax_jit_func=jax.checkpoint, kwargs_for_jax=kwargs)



import functools

import jax
import jax.numpy as jnp

from jax.experimental.pallas.ops.tpu.splash_attention import splash_attention_kernel
from jax.experimental.pallas.ops.tpu.splash_attention import splash_attention_mask
from jax.experimental.shard_map import shard_map


def tpu_splash_attention(
    mesh,
    q_sharding,
    # Input should be of shape (batch, length, heads, kv_dim)
    apply_shard_map,
    query: jax.Array,
    key: jax.Array,
    value: jax.Array,
    decoder_segment_ids: jax.Array | None,
    attn_logits_soft_cap: float | None = None,
) -> jax.Array:
  """TPU Flash Attention."""
  if decoder_segment_ids is not None:
    decoder_segment_ids = splash_attention_kernel.SegmentIds(
        decoder_segment_ids, decoder_segment_ids)

  print('HERE', locals())

  global_block_q = 1024
  global_block_kv = 512
  global_block_kv_compute = 512
  global_block_q_dkv = 2048
  global_block_kv_dkv = 512
  global_block_kv_dkv_compute = 512
  global_block_q_dq = 2048
  global_block_kv_dq = 512
  global_use_fused_bwd_kernel = False
  global_q_layout = 'HEAD_DIM_MINOR'
  global_k_layout = 'HEAD_DIM_MINOR'
  global_v_layout = 'HEAD_DIM_MINOR'

  def wrap_flash_attention(query, key, value, decoder_segment_ids):
    if decoder_segment_ids is not None:
      assert (
          query.shape[2] == decoder_segment_ids.q.shape[1]
      ), "Sharding along sequence dimension not allowed in tpu kernel attention"
    block_sizes = splash_attention_kernel.BlockSizes(
        block_q=min(global_block_q, query.shape[2]),
        block_kv=min(global_block_kv, key.shape[2]),
        block_kv_compute=min(global_block_kv_compute, key.shape[2]),
        block_q_dkv=min(global_block_q_dkv, query.shape[2]),
        block_kv_dkv=min(global_block_kv_dkv, key.shape[2]),
        block_kv_dkv_compute=min(global_block_kv_dkv_compute, query.shape[2]),
        block_q_dq=None if global_use_fused_bwd_kernel else min(
            global_block_q_dq, query.shape[2]),
        block_kv_dq=None if global_use_fused_bwd_kernel else min(
            global_block_kv_dq, query.shape[2]),
        use_fused_bwd_kernel=global_use_fused_bwd_kernel,
        q_layout=splash_attention_kernel.QKVLayout[global_q_layout],
        k_layout=splash_attention_kernel.QKVLayout[global_k_layout],
        v_layout=splash_attention_kernel.QKVLayout[global_v_layout],
    )

    mask = splash_attention_mask.CausalMask(
        shape=(query.shape[2], query.shape[2]))

    # Create multi-head mask
    multi_head_mask = splash_attention_mask.MultiHeadMask(
        masks=(mask,) * query.shape[1])
    #splash_kernel = splash_attention_kernel.make_splash_mha(
    splash_kernel = splash_attention_kernel.make_splash_mha(
        mask=multi_head_mask,
        head_shards=1,
        q_seq_shards=1,
        block_sizes=block_sizes,
        attn_logits_soft_cap=attn_logits_soft_cap,
    )

    return jax.vmap(splash_kernel)(
        query, key, value, segment_ids=decoder_segment_ids)

  if apply_shard_map:
    wrap_flash_attention = shard_map(
        wrap_flash_attention,
        mesh=mesh,
        in_specs=(
            q_sharding,
            q_sharding,
            q_sharding,
            None,
        ),
        out_specs=q_sharding,
        check_rep=False,
    )

  x = wrap_flash_attention(query, key, value, decoder_segment_ids)
  return x


if __name__ == '__main__':
  main()



from typing import Any, Callable


def partition(original: list[Any],
              func: Callable[[Any], bool]) -> tuple[list[Any], list[Any]]:
  """Partitions elements into two parallel lists based on a predicate function.

  Iterates through the 'original' list, applying 'func' to each element 'a'.
  - If `func(a)` returns True, 'a' is appended to the first list ('truthy')
    and `None` is appended to the second list ('falsy').
  - If `func(a)` returns False, `None` is appended to the first list ('truthy')
    and 'a' is appended to the second list ('falsy').

  The result is two lists of the same length as the 'original' list, acting
  as parallel representations of the partitioned elements, using `None` as
  placeholders.

  This is useful when we want to mark a group of elements as static (via passing
  static_argnums) or donated (via donate_argnums) when combining with jax.jit
  and friends.

  Args:
      original: The list of elements to partition.
      func: A callable (function or lambda) that accepts an element from
            'original' and returns a boolean value (True or False).

  Returns:
      A tuple containing two lists (`truthy`, `falsy`), both of the same
      length as `original`:
      - The first list contains elements `x` where `func(x)` was True, and
        `None` otherwise.
      - The second list contains elements `x` where `func(x)` was False, and
        `None` otherwise.

  Example:
      >>> def is_even(n): return n % 2 == 0
      >>> nums = [1, 2, 3, 4, 5, 6]
      >>> truthy_list, falsy_list = partition(nums, is_even)
      >>> truthy_list
      [None, 2, None, 4, None, 6]
      >>> falsy_list
      [1, None, 3, None, 5, None]
  """
  truthy = []
  falsy = []
  for a in original:
    t, f = (a, None) if func(a) else (None, a)
    truthy.append(t)
    falsy.append(f)
  return truthy, falsy


def merge(list1: list[Any], list2: list[Any]) -> list[Any]:
  """Merges two lists element-wise, prioritizing non-None elements from list1.

  Creates a new list where each element is taken from the corresponding position
  in 'list1', unless that element is None, in which case the element from the
  corresponding position in 'list2' is used. Assumes both lists have the
  same length.

  Invariant: merge(*partion(input_list, predicate)) == input_list for any predicate

  Args:
      list1: The primary list. Its elements are preferred unless they are None.
      list2: The secondary list. Its elements are used as fallbacks when the
              corresponding element in list1 is None.

  Returns:
      A new list representing the merged result.

  Raises:
      AssertionError: If 'list1' and 'list2' do not have the same length.

  Example:
      >>> l1 = [1, None, 3, None]
      >>> l2 = [None, 2, None, 4]
      >>> merge(l1, l2)
      [1, 2, 3, 4]
      >>> l3 = [None, 'b', None]
      >>> l4 = ['a', None, 'c']
      >>> merge(l3, l4)
      ['a', 'b', 'c']
  """
  assert len(list1) == len(list2)
  res = []
  for a, b in zip(list1, list2):
    res.append(b if a is None else a)
  return res



import torch
import torch.utils._pytree as torch_pytree
import jax
from enum import Enum
from typing import Union, List, Tuple, Optional, Any, cast
from abc import ABC, abstractmethod

# Reference to original PyTorch native functions
# https://github.com/pytorch/pytorch/blob/main/aten/src/ATen/native/native_functions.yaml


class ViewInfoType(Enum):
  INVALID = 0
  NARROW = 1
  NO_OP = 2
  PERMUTE = 3
  RESHAPE = 4
  RESIZE = 5
  SELECT = 6
  AS_STRIDED = 7
  DIAGONAL = 8


class ViewInfo(ABC):
  """
    Abstract base class for all view operations.
    Defines the interface for applying and updating view transformations.
    """

  def __init__(
      self,
      view_info_type: ViewInfoType = ViewInfoType.INVALID,
  ):
    """
        Initialize a ViewInfo object.

        Args:
            view_info_type: The type of view operation
        """
    self.view_info_type = view_info_type

  @abstractmethod
  def update_tensor(self, new_value: jax.Array,
                    jax_array: jax.Array) -> jax.Array:
    """
        Apply this view transformation to a JAX array and update its value.

        Args:
            new_value: The new values to set in the view
            jax_array: The parent array to update

        Returns:
            Updated array
        """
    pass

  @abstractmethod
  def transform_tensor(self, jax_array: jax.Array) -> jax.Array:
    """
        Apply this view transformation to a JAX array.

        Args:
            jax_array: The array to transform

        Returns:
            Transformed array
        """
    pass

  @abstractmethod
  def calculate_output_shape(self, source: jax.Array) -> List[int]:
    """
        Calculate the resulting shape after applying this view.

        Args:
            source: Original jax array before transformation

        Returns:
            Resulting shape after transformation
        """
    pass


class NarrowInfo(ViewInfo):
  """
    Represents a slicing operation on a tensor.
    Handles operations like tensor[1:3, :, 2:5:2].
    """

  def __init__(self, slices: Union[slice, Tuple[slice]]) -> None:
    """
        Args:
            slices: The slice(s) to apply to the tensor.
                E.g. jax_array.at[slices] will return the transformed tensor.
        """
    super().__init__(ViewInfoType.NARROW)
    self.slices = slices

  def __eq__(self, other: object) -> bool:
    if not isinstance(other, NarrowInfo):
      return False
    return self.slices == other.slices

  def transform_tensor(self, jax_array: jax.Array) -> jax.Array:
    try:
      return jax_array[self.slices]
    except IndexError as e:
      raise IndexError("Invalid slice operation") from e

  def update_tensor(self, new_value: jax.Array,
                    jax_array: jax.Array) -> jax.Array:
    return jax_array.at[self.slices].set(new_value)

  def calculate_output_shape(self, source: jax.Array) -> List[int]:
    return source[self.slices].shape


class SelectInfo(ViewInfo):
  """
    Represents a selection operation on a tensor.
    Typically used for indexing operations that select specific elements.
    """

  def __init__(self,
               dim: int = 0,
               start: int = 0,
               end: int = 0,
               stride: int = 0) -> None:
    super().__init__(ViewInfoType.SELECT)
    self.dim: int = dim
    self.start: int = start
    self.end: int = end
    self.stride: int = stride

  def __eq__(self, other: object) -> bool:
    if not isinstance(other, SelectInfo):
      return False
    return (self.dim == other.dim and self.start == other.start and
            self.end == other.end and self.stride == other.stride)

  def transform_tensor(self, jax_array: jax.Array) -> jax.Array:
    raise NotImplementedError("SelectInfo.apply not implemented")

  def update_tensor(self, new_value: jax.Array,
                    jax_array: jax.Array) -> jax.Array:
    raise NotImplementedError("SelectInfo.update not implemented")

  def calculate_output_shape(self, source: jax.Array) -> List[int]:
    raise NotImplementedError(
        "SelectInfo.calculate_output_shape not implemented")


class AsStridedInfo(ViewInfo):
  """
    Information for as_strided operations.
    """

  def __init__(self, stride: List[int], offset: int = 0) -> None:
    super().__init__(ViewInfoType.AS_STRIDED)
    self.stride: List[int] = stride
    self.offset: int = offset

  def __eq__(self, other: object) -> bool:
    if not isinstance(other, AsStridedInfo):
      return False
    return self.offset == other.offset and self.stride == other.stride

  def transform_tensor(self, jax_array: jax.Array) -> jax.Array:
    raise NotImplementedError("AsStridedInfo.apply not implemented")

  def update_tensor(self, new_value: jax.Array,
                    jax_array: jax.Array) -> jax.Array:
    raise NotImplementedError("AsStridedInfo.update not implemented")

  def calculate_output_shape(self, source: jax.Array) -> List[int]:
    raise NotImplementedError(
        "AsStridedInfo.calculate_output_shape not implemented")


class DiagonalInfo(ViewInfo):
  """
    Information for diagonal operations.
    Extracts diagonal elements from a tensor.
    """

  def __init__(self, offset: int = 0, dim1: int = 0, dim2: int = 1) -> None:
    """
        Args:
            offset: Offset from the main diagonal
            dim1: First dimension for diagonal extraction
            dim2: Second dimension for diagonal extraction
        """
    super().__init__(ViewInfoType.DIAGONAL)
    self.offset: int = offset
    self.dim1: int = dim1
    self.dim2: int = dim2

  def __eq__(self, other: object) -> bool:
    if not isinstance(other, DiagonalInfo):
      return False
    return (self.offset == other.offset and self.dim1 == other.dim1 and
            self.dim2 == other.dim2)

  def transform_tensor(self, jax_array: jax.Array) -> jax.Array:
    raise NotImplementedError("DiagonalInfo.apply not implemented")

  def update_tensor(self, new_value: jax.Array,
                    jax_array: jax.Array) -> jax.Array:
    raise NotImplementedError("DiagonalInfo.update not implemented")

  def calculate_output_shape(self, source: jax.Array) -> List[int]:
    raise NotImplementedError(
        "DiagonalInfo.calculate_output_shape not implemented")


class View(torch.Tensor):
  """
    A View is a reference to another Tensor or another View,
    with a transformation applied to it.
    """

  @staticmethod
  def __new__(cls, parent: Union["torchax.Tensor", "View"], view_info: ViewInfo,
              env: Any) -> "View":
    """
        Args:
            parent: Parent tensor or view
            view_info: Information about the view transformation
            env: Environment for tensor operations
        """
    shape = view_info.calculate_output_shape(parent.jax())
    return torch.Tensor._make_wrapper_subclass(
        cls,
        shape,
        device="meta",
        dtype=parent.dtype,
        requires_grad=False,
    )

  def __init__(self, parent: Union["torchax.Tensor", "View"],
               view_info: ViewInfo, env: Any) -> None:
    super().__init__()
    self.parent = parent
    self.view_info = view_info
    self._env = env

  def get_transformation_chain(self) -> List[ViewInfo]:
    """
        Get all view transformations from the source tensor to this view.
        """
    if isinstance(self.parent, View):
      transformations = self.parent.get_transformation_chain()
      transformations.append(self.view_info)
      return transformations
    else:
      return [self.view_info]

  __torch_function__ = torch._C._disabled_torch_function_impl

  def source_jax(self) -> jax.Array:
    """
        Returns the source tensor.
        """
    if isinstance(self.parent, View):
      return self.parent.source_jax()
    else:
      return self.parent.jax()

  def replace_source_jax(self, new_value: jax.Array) -> None:
    """
        Update the source tensor with new values.
        """
    if isinstance(self.parent, View):
      self.parent.replace_source_jax(new_value)
    else:
      assert new_value.shape == self.parent._elem.shape
      self.parent._elem = new_value

  def torch(self) -> "torchax.Tensor":
    """
        Returns a Torchax tensor representing this view after all transformations
        """
    from torchax.tensor import Tensor

    return Tensor(self.jax(), self._env)

  def update(
      self,
      new_values: Union[jax.Array, "View", "torchax.Tensor"],
      view_infos: Optional[List[ViewInfo]] = None,
  ) -> None:
    """
        Update this view with new values, propagating changes back to source.
        If view_infos is None, it will use the transformation chain
        from the source tensor.
        """
    if view_infos is None:
      view_infos = self.get_transformation_chain()

    # Get the source JAX array
    source_array = self.source_jax()

    # Get the new value
    from torchax.tensor import Tensor

    if isinstance(new_values, View) or isinstance(new_values, Tensor):
      new_values = new_values.jax()

    # Apply all view transformations to the source array
    # And store intermediate values
    intermediate_values = [source_array]
    for view_info in view_infos[:-1]:
      intermediate_values.append(
          view_info.transform_tensor(intermediate_values[-1]))

    # TODO: Investigate efficiency of this algorithm
    # Update the source array with the new value by
    # applying inverse transformations in reverse order
    for view_info, parent_array in zip(
        reversed(view_infos), reversed(intermediate_values)):
      # Apply the inverse transformation to propagate changes back
      new_values = view_info.update_tensor(new_values, parent_array)

    # Update the source tensor with the new values
    self.replace_source_jax(new_values)

  @classmethod
  def __torch_dispatch__(
      cls,
      func: Any,
      types: Tuple[Any, ...],
      args: Tuple[Any, ...] = (),
      kwargs: Optional[dict] = None,
  ) -> Any:
    raise AssertionError(
        'torchax Tensors can only do math within the torchax environment.'
        'Please wrap your code with `with torchax.default_env()` or '
        'call torchax.enable_globally() before.')

  def create_sub_view(self, view_info: ViewInfo) -> "View":
    """
        Create a new view that is a child of this view.
        """
    return View(self, view_info, self._env)

  def __str__(self) -> str:
    return f"View({self.torch()})"

  def jax(self) -> jax.Array:
    """
        Returns a copy of the source tensor after transformations.
        """
    result = self.source_jax()
    for view_info in self.get_transformation_chain():
      result = view_info.transform_tensor(result)
    return result

  def __setitem__(self, indexes, val):
    view_infos = self.get_transformation_chain() + [NarrowInfo(indexes)]
    self.update(view_infos=view_infos, new_values=val)

  def dim(self):
    return self.ndim

  @property
  def device(self):
    return torch.device("jax:0")

  @property
  def jax_device(self):
    return self.jax().device

  @property
  def ndim(self):
    return len(self.shape)

  __repr__ = __str__



# pylint: disable
import os
from typing import Any, Tuple

from jax.experimental import jax2tf
import tensorflow as tf
import torch
from torchax import export


def exported_program_to_tf_function(ep, enable_xla=True):
  weights, jax_program = export.exported_program_to_jax(ep)
  wrapped = lambda *args: jax_program(weights, (args,))
  avals = export.extract_avals(ep)
  input_signature = [
      tf.TensorSpec(shape=t.shape, dtype=t.dtype, name=f"args_{i}")
      for i, t in enumerate(avals)
  ]
  tf_f = tf.function(
      jax2tf.convert(
          wrapped,
          with_gradient=False,
          enable_xla=enable_xla,
      ),
      autograph=False,
      input_signature=input_signature,
  )
  return tf_f


def exported_program_to_tf_module(ep: torch.export.ExportedProgram,
                                  enable_xla=True) -> tf.Module:
  tfm = tf.Module()
  tfm.f = exported_program_to_tf_function(ep, enable_xla)
  return tfm


def save_exported_program_as_tf_saved_model(
    ep: torch.export.ExportedProgram,
    saved_model_dir: os.PathLike,
    serving_key: str = tf.saved_model.DEFAULT_SERVING_SIGNATURE_DEF_KEY,
    function_alias: str = "",
    enable_xla=True,
):
  """This function will export and save a pytorch ExportedProgram to tf.saved_model format.

  The resulting tf.saved_model can be used inference using tf.serving model
  server
  or further convert to tflite flatbuffer for on-device serving.

  Args:
    torch_model: torch.nn.Module - model to export and save
    args: Tuple[Any] - a set of args to trace the model with, i.e.
      torch_model(*args) must run
    saved_model_dir: os.PathLike - location to an empty directory to store the
      saved_model
    serving_key: str  - serving key tag, this is used by tf.serving to know
      which function to run.
    function_alias: str - passed through saved_model.save, used to tag a
      function for inference converter or other tools.
  """
  tfm = exported_program_to_tf_module(ep, enable_xla=enable_xla)
  signatures = {
      serving_key: tfm.f.get_concrete_function(*tfm.f.input_signature)
  }
  save_options = tf.saved_model.SaveOptions(function_aliases={
      function_alias: tfm.f,
  })
  tf.saved_model.save(
      tfm,
      saved_model_dir,
      signatures=signatures,
      options=save_options,
  )


def save_torch_module_as_tf_saved_model(
    torch_model: torch.nn.Module,
    args: Tuple[Any],
    saved_model_dir: os.PathLike,
    serving_key: str = tf.saved_model.DEFAULT_SERVING_SIGNATURE_DEF_KEY,
    function_alias: str = "",
    enable_xla=True,
):
  """This function will export and save a pytorch nn.Module to tf.saved_model format.

  The resulting tf.saved_model can be used inference using tf.serving model
  server
  or further convert to tflite flatbuffer for on-device serving.

  Args:
    torch_model: torch.nn.Module - model to export and save
    args: Tuple[Any] - a set of args to trace the model with, i.e.
      torch_model(*args) must run
    saved_model_dir: os.PathLike - location to an empty directory to store the
      saved_model
    serving_key: str  - serving key tag, this is used by tf.serving to know
      which function to run.
    function_alias: str - passed through saved_model.save, used to tag a
      function for inference converter or other tools.
  """
  ep = torch.export.export(torch_model, args)
  save_exported_program_as_tf_saved_model(ep, saved_model_dir, serving_key,
                                          function_alias, enable_xla)


def exported_program_to_tflite_flatbuffer(ep: torch.export.ExportedProgram):
  tfm = exported_program_to_tf_module(ep)
  tf_concrete_func = tfm.f.get_concrete_function(*tfm.f.input_signature)
  converter = tf.lite.TFLiteConverter.from_concrete_functions(
      [tf_concrete_func], tfm)
  tflite_model = converter.convert()
  return tflite_model


def torch_module_to_tflite_flatbuffer(torch_model: torch.nn.Module,
                                      args: Tuple[Any]):
  ep = torch.export.export(torch_model, args)
  return exported_program_to_tflite_flatbuffer(ep)



# From: https://github.com/google/gemma_pytorch/blob/main/gemma/config.py

# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Gemma model config."""

import dataclasses
import immutabledict
import torch
from typing import Optional

# Keep a mapping from dtype strings to the supported torch dtypes.
_STR_DTYPE_TO_TORCH_DTYPE = immutabledict.immutabledict({
    'float16': torch.float16,
    'float': torch.float32,
    'float32': torch.float32,
    'bfloat16': torch.bfloat16,
})


@dataclasses.dataclass
class GemmaConfig:
  # The number of tokens in the vocabulary.
  vocab_size: int = 256000
  # The maximum sequence length that this model might ever be used with.
  max_position_embeddings: int = 8192
  # The number of blocks in the model.
  num_hidden_layers: int = 28
  # The number of attention heads used in the attention layers of the model.
  num_attention_heads: int = 16
  # The number of key-value heads for implementing attention.
  num_key_value_heads: int = 16
  # The hidden size of the model.
  hidden_size: int = 3072
  # The dimension of the MLP representations.
  intermediate_size: int = 24576
  # The number of head dimensions.
  head_dim: int = 256
  # The epsilon used by the rms normalization layers.
  rms_norm_eps: float = 1e-6
  # The dtype of the weights.
  dtype: str = 'bfloat16'
  # Whether a quantized version of the model is used.
  quant: bool = False
  # The path to the model tokenizer.
  tokenizer: Optional[str] = 'tokenizer/tokenizer.model'

  def get_dtype(self) -> Optional[torch.dtype]:
    """Gets the torch dtype from the config dtype string."""
    return _STR_DTYPE_TO_TORCH_DTYPE.get(self.dtype, None)


def get_config_for_7b() -> GemmaConfig:
  return GemmaConfig()


def get_config_for_2b() -> GemmaConfig:
  return GemmaConfig(
      num_hidden_layers=18,
      num_attention_heads=8,
      num_key_value_heads=1,
      hidden_size=2048,
      intermediate_size=16384)


def get_model_config(variant: str) -> GemmaConfig:
  if variant == '7b':
    return get_config_for_7b()
  elif variant == '2b':
    return get_config_for_2b()
  return ValueError(f'Invalid variant {variant}. Supported variants are "2b"'
                    'and "7b"')



import torch
import torchax
import re
import sys
import unittest

from torchax.tensor import Tensor
from torchax.view import View


class TrainTest(unittest.TestCase):

  def setUp(self):
    torch.manual_seed(0)
    torchax.enable_globally()

  def test_copy_(self):
    x = torch.zeros((10, 10), device="jax")
    y = torch.ones((5, 5), device="jax")
    x[0:5, :][:, 0:5].copy_(y[:, :])
    self.assertEqual(type(x), Tensor)
    self.assertEqual(x.shape, (10, 10))
    self.assertEqual(x[0:5, 0:5].sum(), 25)
    self.assertEqual(x.sum(), 25)

  def test_transivity(self):
    x = torch.zeros((10, 10), device="jax")
    x_view = x[0:5, :][:, 0:5].add_(1)
    y_view = x_view[0:5, :][:, 0:5].add_(1)
    self.assertEqual(type(x), Tensor)
    self.assertEqual(type(x_view), View)
    self.assertEqual(type(y_view), View)
    self.assertEqual(x.shape, (10, 10))
    self.assertEqual(x[0:5, 0:5].sum(), 50)
    self.assertEqual(x.sum(), 50)

  def test_outofplace_add(self):
    x = torch.zeros((10, 10), device="jax")
    x2 = x[0:5, :][:, 0:5].add(1)
    x3 = x2[0:5, :][:, 0:5].add_(x2)
    self.assertEqual(type(x), Tensor)
    self.assertEqual(type(x2), Tensor)
    self.assertEqual(type(x3), View)
    self.assertEqual(x.shape, (10, 10))
    self.assertEqual(x.sum(), 0)
    self.assertEqual(x2.sum(), 50)

  def test_multiply_tensor_and_view(self):
    x = torch.ones((10, 10), device="jax") * 2
    y = torch.ones((10, 10), device="jax")
    x1 = x[:, :]
    res = x1.mul(y)
    self.assertEqual(type(x), Tensor)
    self.assertEqual(type(y), Tensor)
    self.assertEqual(type(x1), View)
    self.assertEqual(type(res), Tensor)
    self.assertEqual(res.sum(), 200)

  def test_multiply_views(self):
    x = torch.ones((10, 10), device="jax") * 2
    y = torch.ones((10, 10), device="jax")
    x1 = x[0:1, :]
    y1 = y[0:1, :]
    res = x1.mul(y1)
    self.assertEqual(type(x), Tensor)
    self.assertEqual(type(y), Tensor)
    self.assertEqual(type(x1), View)
    self.assertEqual(type(y1), View)
    self.assertEqual(type(res), Tensor)
    self.assertEqual(res.sum(), 20)

  def test_setitem(self):
    a = torch.zeros(10, device="jax")
    a[0:5][0:3] = 1
    self.assertEqual(type(a), Tensor)
    self.assertEqual(a.shape, (10,))
    self.assertEqual(a.sum(), 3)

  # Test all in-place operations
  def test_add_(self):
    x = torch.zeros((10, 10), device="jax")
    x[0:5, :][:, 0:5].add_(1)
    self.assertEqual(type(x), Tensor)
    self.assertEqual(x.shape, (10, 10))
    self.assertEqual(x.sum(), 25)

  def test_sub_(self):
    x = torch.zeros((10, 10), device="jax")
    x[0:5, :][:, 0:5].sub_(1)
    self.assertEqual(type(x), Tensor)
    self.assertEqual(x.shape, (10, 10))
    self.assertEqual(x.sum(), -25)

  def test_mul_(self):
    x = torch.ones((10, 10), device="jax")
    x[0:5, :][:, 0:5].mul_(2)
    self.assertEqual(type(x), Tensor)
    self.assertEqual(x.shape, (10, 10))
    self.assertEqual(x.sum(), 125)

  def test_div_(self):
    x = torch.ones((10, 10), device="jax")
    x[0:10, :][:, 0:10].div_(2)
    self.assertEqual(type(x), Tensor)
    self.assertEqual(x.shape, (10, 10))
    self.assertEqual(x.sum(), 50)

  def test_pow_(self):
    x = torch.full((10, 10), fill_value=2, device="jax")
    x[0:5, :][:, 0:5].pow_(2)
    self.assertEqual(type(x), Tensor)
    self.assertEqual(x.shape, (10, 10))
    self.assertEqual(x.sum(), 250)

  def test_clamp_(self):
    x = torch.arange(100, device="jax", dtype=torch.float).reshape(10, 10)
    x[0:5, :][:, 0:5].clamp_(min=50, max=80)
    self.assertEqual(type(x), Tensor)
    self.assertEqual(x.shape, (10, 10))
    self.assertTrue((x[0:5, 0:5] >= 50).all())
    self.assertTrue((x[0:5, 0:5] <= 80).all())

  def test_lt_(self):
    x = torch.ones((10, 10), device="jax")
    y = torch.zeros((10, 10), device="jax")
    x[0:5, :][:, 0:5].lt_(0.5)
    self.assertEqual(type(x), Tensor)
    self.assertEqual(x.shape, (10, 10))
    self.assertEqual(x[0:5, 0:5].sum(),
                     0)  # All False (0) in the modified region
    self.assertEqual(x[5:, 5:].sum(),
                     25)  # All True (1) in the unmodified region

  def test_le_(self):
    x = torch.ones((10, 10), device="jax")
    x[0:5, :][:, 0:5].le_(1)
    self.assertEqual(type(x), Tensor)
    self.assertEqual(x.shape, (10, 10))
    self.assertEqual(x.sum(), 100)  # All True (1)

  def test_gt_(self):
    x = torch.ones((10, 10), device="jax")
    x[0:5, :][:, 0:5].gt_(1)
    self.assertEqual(type(x), Tensor)
    self.assertEqual(x.shape, (10, 10))
    self.assertEqual(x[0:5, 0:5].sum(),
                     0)  # All False (0) in the modified region
    self.assertEqual(x.sum(), 75)  # Only the unmodified region is True (1)

  def test_ge_(self):
    x = torch.ones((10, 10), device="jax")
    x[0:5, :][:, 0:5].ge_(1)
    self.assertEqual(type(x), Tensor)
    self.assertEqual(x.shape, (10, 10))
    self.assertEqual(x.sum(), 100)  # All True (1)

  def test_eq_(self):
    x = torch.ones((10, 10), device="jax")
    x[0:5, :][:, 0:5].eq_(1)
    self.assertEqual(type(x), Tensor)
    self.assertEqual(x.shape, (10, 10))
    self.assertEqual(x.sum(), 100)  # All True (1)

  def test_ne_(self):
    x = torch.ones((10, 10), device="jax")
    x[0:5, :][:, 0:5].ne_(1)
    self.assertEqual(type(x), Tensor)
    self.assertEqual(x.shape, (10, 10))
    self.assertEqual(x[0:5, 0:5].sum(),
                     0)  # All False (0) in the modified region
    self.assertEqual(x.sum(), 75)  # Only the unmodified region is True (1)

  def test_bernoulli_(self):
    # Set a fixed seed for deterministic behavior
    torch.manual_seed(42)
    x = torch.full((10, 10), fill_value=0.5, device="jax")
    y = x[0:5, :][:, 0:5]
    y.bernoulli_()
    self.assertEqual(type(x), Tensor)
    self.assertEqual(x.shape, (10, 10))
    # Values will be 0 or 1 in the modified region
    self.assertTrue(((x[0:5, 0:5] == 0) | (x[0:5, 0:5] == 1)).all())
    # Unmodified region remains 0.5
    self.assertTrue((x[5:, 5:] == 0.5).all())

  def test_geometric_(self):
    torch.manual_seed(42)
    x = torch.full((10, 10), fill_value=0.5, device="jax")
    y = x[0:5, :][:, 0:5]
    y.geometric_(p=0.5)
    self.assertEqual(type(x), Tensor)
    self.assertEqual(x.shape, (10, 10))
    # Geometric distribution values are positive integers
    self.assertTrue((x[0:5, 0:5] >= 1).all())
    # Unmodified region remains 0.5
    self.assertTrue((x[5:, 5:] == 0.5).all())

  def test_normal_(self):
    torch.manual_seed(42)
    x = torch.zeros((10, 10), device="jax")
    x[0:5, :][:, 0:5].normal_(mean=0, std=1)
    self.assertEqual(type(x), Tensor)
    self.assertEqual(x.shape, (10, 10))
    # Unmodified region remains 0
    self.assertEqual(x[5:, 5:].sum(), 0)

  def test_uniform_(self):
    torch.manual_seed(42)
    x = torch.zeros((10, 10), device="jax")
    x[0:5, :][:, 0:5].uniform_(0, 1)
    self.assertEqual(type(x), Tensor)
    self.assertEqual(x.shape, (10, 10))
    # Values in modified region are between 0 and 1
    self.assertTrue((x[0:5, 0:5] >= 0).all())
    self.assertTrue((x[0:5, 0:5] <= 1).all())
    # Unmodified region remains 0
    self.assertEqual(x[5:, 5:].sum(), 0)

  def test_relu_(self):
    x = torch.randn((10, 10), device="jax")
    x_copy = x.clone()
    x[0:5, :][:, 0:5].relu_()
    self.assertEqual(type(x), Tensor)
    self.assertEqual(x.shape, (10, 10))
    # Modified region has no negative values
    self.assertTrue((x[0:5, 0:5] >= 0).all())
    # Unmodified region remains the same
    self.assertTrue(torch.equal(x[5:, 5:], x_copy[5:, 5:]))

  def test_squeeze_(self):
    x = torch.randn((10, 1, 10), device="jax")
    x_clone = x.clone()
    # Squeeze the middle dimension
    x.squeeze_(1)
    self.assertEqual(type(x), Tensor)
    self.assertEqual(x.shape, (10, 10))
    # Content should remain the same
    self.assertTrue(torch.allclose(x, x_clone.squeeze()))

  def test_sqrt_(self):
    x = torch.randn((10, 10),
                    device="jax").abs()  # Use abs to ensure positive values
    x_copy = x.clone()
    x[0:5, :][:, 0:5].sqrt_()
    self.assertEqual(type(x), Tensor)
    self.assertEqual(x.shape, (10, 10))
    # Modified region is square root of original values
    self.assertTrue(torch.allclose(x[0:5, 0:5], torch.sqrt(x_copy[0:5, 0:5])))
    # Unmodified region remains the same
    self.assertTrue(torch.equal(x[5:, 5:], x_copy[5:, 5:]))

  def test_clamp_min_(self):
    x = torch.randn((10, 10), device="jax")
    x_copy = x.clone()
    x[0:5, :][:, 0:5].clamp_min_(0)
    self.assertEqual(type(x), Tensor)
    self.assertEqual(x.shape, (10, 10))
    # Modified region has no values below 0
    self.assertTrue((x[0:5, 0:5] >= 0).all())
    # Unmodified region remains the same
    self.assertTrue(torch.equal(x[5:, 5:], x_copy[5:, 5:]))

  def test_sigmoid_(self):
    x = torch.randn((10, 10), device="jax")
    x_copy = x.clone()
    x[0:5, :][:, 0:5].sigmoid_()
    self.assertEqual(type(x), Tensor)
    self.assertEqual(x.shape, (10, 10))
    # Modified region values are between 0 and 1
    self.assertTrue((x[0:5, 0:5] >= 0).all())
    self.assertTrue((x[0:5, 0:5] <= 1).all())
    # Unmodified region remains the same
    self.assertTrue(torch.equal(x[5:, 5:], x_copy[5:, 5:]))

  def test_tanh_(self):
    x = torch.randn((10, 10), device="jax")
    x_copy = x.clone()
    x[0:5, :][:, 0:5].tanh_()
    self.assertEqual(type(x), Tensor)
    self.assertEqual(x.shape, (10, 10))
    # Modified region values are between -1 and 1
    self.assertTrue((x[0:5, 0:5] >= -1).all())
    self.assertTrue((x[0:5, 0:5] <= 1).all())
    # Unmodified region remains the same
    self.assertTrue(torch.equal(x[5:, 5:], x_copy[5:, 5:]))

  def test_ceil_(self):
    x = torch.randn((10, 10), device="jax")
    x_copy = x.clone()
    x[0:5, :][:, 0:5].ceil_()
    self.assertEqual(type(x), Tensor)
    self.assertEqual(x.shape, (10, 10))
    # Check that ceil operation was applied correctly
    self.assertTrue(torch.allclose(x[0:5, 0:5], torch.ceil(x_copy[0:5, 0:5])))
    # Unmodified region remains the same
    self.assertTrue(torch.equal(x[5:, 5:], x_copy[5:, 5:]))

  def test_logical_not_(self):
    x = torch.zeros((10, 10), device="jax")
    x[0:5, 0:5] = 1  # Set some values to 1
    x[0:5, :][:, 0:5].logical_not_()
    self.assertEqual(type(x), Tensor)
    self.assertEqual(x.shape, (10, 10))
    # Modified region has all values flipped
    self.assertEqual(x[0:5, 0:5].sum(), 0)  # All now 0
    # Unmodified region remains 0
    self.assertEqual(x[5:, 5:].sum(), 0)

  def test_unsqueeze_(self):
    x = torch.randn((10, 10), device="jax")
    x_copy = x.clone()
    # Add dimension at index 1
    x.unsqueeze_(1)
    self.assertEqual(type(x), Tensor)
    self.assertEqual(x.shape, (10, 1, 10))
    # Content should remain the same
    self.assertTrue(torch.equal(x.squeeze(1), x_copy))

  def test_transpose_(self):
    x = torch.randn((10, 5), device="jax")
    x_copy = x.clone()
    # Transpose dimensions 0 and 1
    x.transpose_(0, 1)
    self.assertEqual(type(x), Tensor)
    self.assertEqual(x.shape, (5, 10))
    # Check transposition worked correctly
    self.assertTrue(torch.equal(x, x_copy.transpose(0, 1)))

  def test_log_normal_(self):
    torch.manual_seed(42)
    x = torch.zeros((10, 10), device="jax")
    x[0:5, :][:, 0:5].log_normal_(mean=0, std=1)
    self.assertEqual(type(x), Tensor)
    self.assertEqual(x.shape, (10, 10))
    # Log-normal values are positive
    self.assertTrue((x[0:5, 0:5] > 0).all())
    # Unmodified region remains 0
    self.assertEqual(x[5:, 5:].sum(), 0)

  def test_scatter_add_(self):
    # Initialize test tensors
    x = torch.zeros((5, 5), device="jax")
    indices = torch.tensor([[0, 1, 2], [0, 1, 2]], device="jax")
    values = torch.ones((2, 3), device="jax")

    # Apply scatter_add_ operation
    x.scatter_add_(0, indices, values)

    self.assertEqual(type(x), Tensor)
    self.assertEqual(x.shape, (5, 5))
    # Check specific values were added
    self.assertTrue(torch.all(x[0, 0] == 2.0))
    self.assertEqual(x.sum(), 6.0)  # Only the 3 specified positions have values

  def test_scatter_(self):
    # Initialize test tensors
    x = torch.zeros((5, 5), device="jax")
    indices = torch.tensor([[0, 1, 2], [0, 1, 2]], device="jax")
    values = torch.ones((2, 3), device="jax") * 2.0

    # Apply scatter_ operation
    x.scatter_(0, indices, values)

    self.assertEqual(type(x), Tensor)
    self.assertEqual(x.shape, (5, 5))
    # Check specific values were replaced
    self.assertEqual(x[0, 0], 2.0)
    self.assertEqual(x[1, 1], 2.0)
    self.assertEqual(x[2, 2], 2.0)
    self.assertEqual(x.sum(), 6.0)  # Only the 3 specified positions have values

  def test_scatter_reduce_(self):
    # Initialize test tensors
    x = torch.ones((5, 5), device="jax")
    indices = torch.tensor([[0, 1, 2], [0, 1, 2]], device="jax")
    values = torch.ones((2, 3), device="jax") * 2.0

    # Apply scatter_reduce_ operation with "sum" reduction
    x.scatter_reduce_(0, indices, values, reduce="sum")

    self.assertEqual(type(x), Tensor)
    self.assertEqual(x.shape, (5, 5))
    # Check specific values were reduced
    self.assertTrue(torch.all(x[0, 0] == 5.0))
    self.assertEqual(x.sum(), 37.0)



from typing import Callable
from absl.testing import absltest
from absl.testing import parameterized
import torch
import torchax
import torchax.tensor


class SeqModel(torch.nn.Module):
  """ Architecture is LLM generated """

  def __init__(self):
    super().__init__()
    self.gru = torch.nn.GRU(20, 30, batch_first=True)
    self.linear = torch.nn.Linear(30, 1)

  def forward(self, x: torch.Tensor):
    output, _ = self.gru(x)  #output, hidden state
    output = self.linear(output)
    return output


class TestTorchFunctions(parameterized.TestCase):

  def setUp(self):
    torchax.enable_globally()
    torchax.enable_accuracy_mode()
    self.env = torchax.default_env()

  @parameterized.named_parameters(
      ('tensor_2d', [[0.1, 1.2], [2.2, 3.1], [4.9, 5.2]]),
      ('tensor_1d', [0, 1]), ('tensor_scalar', 3.14159), ('tensor_empty', []),
      ('tensor_dtype', [[0.11111, 0.222222, 0.3333333]], {
          'dtype': torch.float64
      }))
  def test_tensor_constructor(self, arg, kwargs=None):
    kwargs = kwargs or {}
    expected = torch.tensor(arg, **kwargs)

    actual = torch.tensor(arg, device='jax', **kwargs)
    self.assertIsInstance(actual, torchax.tensor.Tensor)

    torch.testing.assert_close(actual.to('cpu'), expected)

  def test_dont_capture_conversion(self):
    t = torch.tensor([1, 2, 3])
    with self.env:
      t2 = self.env.to_xla(t)
      # assert no exceptions

  def test_brackets(self):
    with self.env:
      a = torch.randn((2, 3))
      a[1] = 9
      self.assertEqual(a[1, 0].item(), 9)

  def test_bernoulli_inplace(self):
    with self.env:
      a = torch.randn((2, 3))
      a.bernoulli_(0.4)

  def test_flatten(self):
    with self.env:
      a = torch.randn((2, 3, 4))
      a = a.flatten(0, 1)
      self.assertEqual(tuple(a.shape), (6, 4))

  def test_rnn(self):
    model = SeqModel()
    x = torch.randn((2, 100, 20))
    res = model(x)
    with self.env:
      model.to('jax')
      x = x.to('jax')
      res2 = model(x)
      print(res.shape, res2.shape)

      self.assertEqual(res.shape, res2.shape)

  def test_rms_norm(self):
    model = torch.nn.RMSNorm((100, 20))
    x = torch.randn((2, 100, 20))
    res = model(x)

    with self.env:
      model.to('jax')
      x = x.to('jax')
      res2 = model(x)
      self.assertTrue(torch.allclose(res, res2.to('cpu')))

  def test_randn_requires_grad(self):
    x = torch.randn((3, 3), requires_grad=True, device='jax')
    self.assertEqual(x.requires_grad, True)


if __name__ == '__main__':
  absltest.main()
